/* ENG is Not a Game (c) Thunor 2008 <thunorsif@hotmail.com>
www.thunor.org.uk

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. */

/* Includes */
#include <SDL/SDL.h>

/* Miscellaneous defines */
#define ENG_UNDEFINED -1
#define ENG_INHERITED -2
#define ENG_PATH_LENGTH 128
#define ENG_FILE_LENGTH 64
#define ENG_TITLE_LENGTH 64
#define ENG_CONFIG_OPTION_NAME_LENGTH 64
#define ENG_CONFIG_OPTION_VALUE_LENGTH 64
#define ENG_CONFIG_OPTION_LENGTH ENG_CONFIG_OPTION_NAME_LENGTH + 1 + ENG_CONFIG_OPTION_VALUE_LENGTH

/* Paths */
#define ENG_COMMON_IMAGE_DIR PACKAGE_DATA_DIR "/images/common/"

/* System images */
#define ENG_IMG_POINTER "imgPointer"
#define ENG_IMG_POINTER_SHADOW "imgPointerShadow"
#define ENG_IMG_POINTER_W 13
#define ENG_IMG_POINTER_H 19
#define ENG_IMG_EXIT "imgExit"
#define ENG_IMG_EXIT_W 24
#define ENG_IMG_EXIT_H 14
#define ENG_IMG_MINIMISE "imgMinimise"
#define ENG_IMG_MINIMISE_W 24
#define ENG_IMG_MINIMISE_H 14
#define ENG_IMG_RESTORE "imgRestore"
#define ENG_IMG_RESTORE_W 24
#define ENG_IMG_RESTORE_H 14
#define ENG_IMG_MAXIMISE "imgMaximise"
#define ENG_IMG_MAXIMISE_W 24
#define ENG_IMG_MAXIMISE_H 14

/* Custom image prefixes */
#define ENG_IMG_TESTCARD "imgTestCard"	/* This will be scaled to the object size in units of 32x32 */
#define ENG_IMG_BLANK "imgBlank"		/* This will be created the same size as the object */

/* System object ids reserved up to 99 */
#define ENG_LYR_ROOT 0
#define ENG_LYR_CANVAS 1
#define ENG_LYR_SCREEN 2
#define ENG_LYR_VIEWPORT 3
#define ENG_LYR_DEBUG 4
#define ENG_LYR_POINTER 9
#define ENG_REC_POINTER 10
#define ENG_FNT_ASCII_8X12_WHITE 20
#define ENG_FNT_ASCII_8X12_BLACK 21
#define ENG_FNT_ASCII_8X12_WHITE_BLACK_OUTLINE 22
#define ENG_FNT_ASCII_16X24_WHITE 23
#define ENG_FNT_ASCII_16X24_BLACK 24
#define ENG_FNT_ASCII_16X24_WHITE_BLACK_OUTLINE 25

/* Object klasses */
#define ENG_OBJECT_KLASSE_RECTANGLE 0
#define ENG_OBJECT_KLASSE_LAYER 1
#define ENG_OBJECT_KLASSE_FONT 2
#define ENG_OBJECT_KLASSE_TEXT 3

/* Object drag points */
#define ENG_DRAG_POINT_ANYWHERE 0
#define ENG_DRAG_POINT_MIDDLE 1

/* Object drag types */
#define ENG_DRAG_ANYWHERE 1
#define ENG_DRAG_RESTRICT_TO_PARENT 2

/* Object sources */
#define ENG_OBJECT_SOURCE_SYSTEM 0
#define ENG_OBJECT_SOURCE_FILE 1
#define ENG_OBJECT_SOURCE_USER 2

/* Colour key */
#define ENG_COLOUR_KEY_RED 255
#define ENG_COLOUR_KEY_GREEN 0
#define ENG_COLOUR_KEY_BLUE 128

/* ZOrder positions */
#define ENG_ZORDER_ABOVE_ALL 0
#define ENG_ZORDER_ABOVE 1
#define ENG_ZORDER_BELOW 2

/* Text related */
#define ENG_TEXT_WRITE_PAGE 0
#define ENG_TEXT_WRITE_LINE 1
#define ENG_TEXT_WRITE_WORD 2

/* Actions */
#define ENG_ACTION_UP 0
#define ENG_ACTION_UP_RIGHT 1
#define ENG_ACTION_RIGHT 2
#define ENG_ACTION_DOWN_RIGHT 3
#define ENG_ACTION_DOWN 4
#define ENG_ACTION_DOWN_LEFT 5
#define ENG_ACTION_LEFT 6
#define ENG_ACTION_UP_LEFT 7
#define ENG_ACTION_SELECT 8
#define ENG_ACTION_SHUTDOWN 9
#define ENG_ACTION_TOGGLE_FULLSCREEN 10
#define ENG_ACTION_MODIFIER1 11
#define ENG_ACTION_MODIFIER2 12
#define ENG_ACTIONS 13

/* Devices */
#define ENG_DEVICE_KEYBOARD 0
#define ENG_DEVICE_JOYSTICK 1
#define ENG_DEVICE_MOUSE 2

/* Events */
#define ENG_EVENT_OVER 10
#define ENG_EVENT_OUT 11
#define ENG_EVENT_PRESS 12
#define ENG_EVENT_CLICK 13
#define ENG_EVENT_RELEASE 14
#define ENG_EVENT_DRAG 15
#define ENG_EVENT_QUITTING 16
#define ENG_EVENT_POST_LOAD_OBJECTS 17
#define ENG_EVENT_PRE_RENDER 18
#define ENG_EVENT_POST_RENDER 19
#define ENG_EVENT_CYCLE 20
#define ENG_EVENT_POST_EFFECT 21
#define ENG_EVENT_UP 30
#define ENG_EVENT_UP_RIGHT 31
#define ENG_EVENT_RIGHT 32
#define ENG_EVENT_DOWN_RIGHT 33
#define ENG_EVENT_DOWN 34
#define ENG_EVENT_DOWN_LEFT 35
#define ENG_EVENT_LEFT 36
#define ENG_EVENT_UP_LEFT 37
#define ENG_EVENT_SELECT 38
#define ENG_EVENT_SHUTDOWN 39
#define ENG_EVENT_TOGGLE_FULLSCREEN 40
#define ENG_EVENT_MODIFIER1 41
#define ENG_EVENT_MODIFIER2 42
//#define ENG_EVENT_SEND_PRE_RENDER
//#define ENG_EVENT_SEND_POST_RENDER
#define ENG_EVENT_SEND_CYCLE

/* Effects */
#define ENG_EFFECT_LINEAR_MOVE 0
#define ENG_EFFECT_ALPHA_FADE 1

/* Effect states */
#define ENG_EFFECT_STATE_PAUSE 0
#define ENG_EFFECT_STATE_UNPAUSE 1
#define ENG_EFFECT_STATE_FINALISE 2
#define ENG_EFFECT_STATE_TERMINATE 3

/* Pointer modes */
#define ENG_POINTER_MODE_OFF 0
#define ENG_POINTER_MODE_PHYSICAL 1
#define ENG_POINTER_MODE_SIMULATED 2
#define ENG_POINTER_MODE_NAVIGATION 3

/* Hit points */
#define ENG_HIT_POINT_TOP_LEFT 0
#define ENG_HIT_POINT_CENTRE 1

/* GP2X button IDs */
#define GP2X_JOY_N 0x00
#define GP2X_JOY_NW 0x01
#define GP2X_JOY_W 0x02
#define GP2X_JOY_SW 0x03
#define GP2X_JOY_S 0x04
#define GP2X_JOY_SE 0x05
#define GP2X_JOY_E 0x06
#define GP2X_JOY_NE 0x07
#define GP2X_START 0x08
#define GP2X_SELECT 0x09
#define GP2X_LTRIG 0x0A
#define GP2X_RTRIG 0x0B
#define GP2X_BTN_A 0x0C
#define GP2X_BTN_B 0x0D
#define GP2X_BTN_X 0x0E
#define GP2X_BTN_Y 0x0F
#define GP2X_VOL_UP 0x10
#define GP2X_VOL_DN 0x11
#define GP2X_BTN_JOY 0x12

/* Zaurus button IDs */
#define ZAURUS_UP SDLK_UP
#define ZAURUS_DOWN SDLK_DOWN
#define ZAURUS_LEFT SDLK_LEFT
#define ZAURUS_RIGHT SDLK_RIGHT
#define ZAURUS_SPACE SDLK_SPACE
#define ZAURUS_CANCEL SDLK_ESCAPE
#define ZAURUS_OK 0
#define ZAURUS_CALENDAR SDLK_F9
#define ZAURUS_ADDRESS SDLK_F10
#define ZAURUS_HOME SDLK_F12
#define ZAURUS_MENU SDLK_F11
#define ZAURUS_EMAIL SDLK_F13

/* Error codes */
#define ENG_ERROR_NONE 0
#define ENG_ERROR_ENG_ALREADY_INITIALISED 10
#define ENG_ERROR_ENG_ALREADY_STARTED 20
#define ENG_ERROR_ENG_NOT_YET_INITIALISED 22
#define ENG_ERROR_ENG_NOT_YET_STARTED 25
#define ENG_ERROR_CANNOT_INIT_SDL 35
#define ENG_ERROR_CANNOT_SET_VIDEO_MODE 40
#define ENG_ERROR_CANNOT_OPEN_OBJECTS_FILE 50
#define ENG_ERROR_OBJECTS_FILE_SYNTAX_ERROR 60
#define ENG_ERROR_TOO_MANY_OBJECTS 70
#define ENG_ERROR_CANNOT_OPEN_IMAGE_FILE 80
#define ENG_ERROR_TOO_MANY_IMAGES 90
#define ENG_ERROR_CANNOT_CREATE_SURFACE 100
#define ENG_ERROR_CANNOT_SET_SURFACE_ALPHA 110
#define ENG_ERROR_CANNOT_FILL_SURFACE 120
#define ENG_ERROR_CANNOT_BLIT_SURFACE 130
#define ENG_ERROR_CANNOT_UPDATE_SURFACE 135
#define ENG_ERROR_INVALID_OP_ENV 140
#define ENG_ERROR_CANNOT_SET_COLOUR_KEY 160
#define ENG_ERROR_CANNOT_TOGGLE_FULLSCREEN 170
#define ENG_ERROR_INVALID_EVENT 200
#define ENG_ERROR_CANNOT_OPEN_JOYSTICK 210
#define ENG_ERROR_INVALID_ZORDER_POSITION 220
#define ENG_ERROR_INVALID_OBJECT 230
#define ENG_ERROR_INVALID_TARGET_OBJECT 240
#define ENG_ERROR_INVALID_IMAGE 260
#define ENG_ERROR_INVALID_EFFECT 280
#define ENG_ERROR_INVALID_EFFECT_ID 290
#define ENG_ERROR_INVALID_EFFECT_STATE 300
#define ENG_ERROR_TOO_MANY_EFFECTS 310
#define ENG_ERROR_CANNOT_OPEN_CONFIG_FILE 320
#define ENG_ERROR_CONFIG_FILE_SYNTAX_ERROR 330
#define ENG_ERROR_TOO_MANY_CONFIG_OPTIONS 340
#define ENG_ERROR_INVALID_CONFIG_OPTION 350
#define ENG_ERROR_CANNOT_WRITE_TO_CONFIG_FILE 360


/* Variables */
typedef struct engobject {
	int id;
	int pid;									/* Once object is loaded, use *pob instead; the engine does not use this after loading */
	int klasse;
	int visible;								/* -2 (inherited), 0 or 1: Invisible objects are not manipulatable */
	int enabled;								/* -2 (inherited), 0 or 1: Disabled objects are not manipulatable */
	int alpha;									/* -2 (inherited) or 0 (transparent) to 255 (opaque) */
	int alpha_over;
	int bgcolour;
	int bgcolour_over;
	int x;										/* Object coordinates and dimensions */
	int y;
	int w;
	int h;
	char img_file[ENG_FILE_LENGTH];				/* Once object is loaded, use *img_surface instead; the engine frees img_file when freeing this object */
	int img_set_colour_key;
	int img_use_rleaccel;
	int img_x;									/* Image region */
	int img_y;
	int img_w;
	int img_h;
	char img_over_file[ENG_FILE_LENGTH];		/* Once object is loaded, use *img_over_surface instead; the engine frees img_over_file when freeing this object */
	int img_over_set_colour_key;
	int img_over_use_rleaccel;
	int img_over_x;								/* Image over region */
	int img_over_y;
	int img_over_w;
	int img_over_h;
	int draggable;
	int drag_point;
	int navigable;								/* Similar to tabbable */
	/* Properties below are set by the system when the object is created */
	int zorder;
	void *pob;									/* Parent object pointer */
	SDL_Surface *img_surface;
	SDL_Surface *img_over_surface;
	int source;									/* The source of this object: system created, file or user created */
	char objects_file[ENG_PATH_LENGTH];			/* If source is file then this is the objects file */
} ENG_Object;

typedef struct engaction {
	int device;					/* Keyboard or joystick */
	int index;					/* The index of the joystick if applicable */
	int id;						/* Key or button id */
	int mod;					/* Modifier action */
	int state;					/* Pressed or released */
	int actionstate;			/* Pressed or released */
	int mx;						/* x coordinate of press if applicable */
	int my;						/* y coordinate of press if applicable */
	ENG_Object *oup;			/* The object under the pointer */
	int cyclesactive;			/* How many cycles this has been pressed for */
	int delay;					/* The initial delay in cycles before repeating - 0 disables the delay */
	int interval;				/* The repeat interval in cycles - 0 disables the interval */
} ENG_Action;

typedef struct engimage {
	char img_file[ENG_FILE_LENGTH];
	SDL_Surface *img_surface;
} ENG_Image;

typedef struct engevent {
	ENG_Object *object;
	int klasse;
} ENG_Event;

typedef struct engopenv {
	char image_dir[ENG_PATH_LENGTH];			/* All images will be found here */
	char objects_file[ENG_PATH_LENGTH];			/* A list of all game objects */
	char config_file[ENG_PATH_LENGTH];			/* The application's config (rc) file */
	int screen_w;								/* The screen/window resolution as in video mode */
	int screen_h;
	int canvas_w;
	int canvas_h;
	int viewport_x;								/* Relative to the canvas */
	int viewport_y;								/* Relative to the canvas */
	int viewport_w;
	int viewport_h;
	int viewport_screen_x;						/* Relative to the screen */
	int viewport_screen_y;						/* Relative to the screen */
	char title[ENG_TITLE_LENGTH];				/* For window managers */
	int fullscreen;								/* True or false */
	int systempointer;							/* True or false */
	int pointer;								/* True or false */
	int pointer_mode;							/* Pointer mode */
	int pointer_move_unit_low;					/* Unit pointer should move unaccelerated */
	int pointer_move_unit_high;					/* Unit pointer should move accelerated */
	int pointer_move_unit_threshold;			/* Number of units marking the low/high threshold */
	int delay;									/* Per main loop cycle (ms) - set to zero if not required */
	int cycle_limit;							/* 16, 20, 25, 33, 50 or 100 - set to zero if not required */
	void (*event_processor)(ENG_Event *event);	/* The function that receives events */
} ENG_OpEnv;

typedef struct engasciitext {
	SDL_Surface *font;			/* Font image surface */
	SDL_Surface *target;		/* Target image surface */
	int img_x;					/* Region to wrap text within */
	int img_y;
	int img_w;
	int img_h;
	int bgcolour;				/* Colour to wipe the surface with prior to writing */
	int char_spacing;			/* A +/- pixel value to adjust the horizontal spacing */
	int line_spacing;			/* A +/- pixel value to adjust the vertical spacing */
	int tab_spaces;				/* The number of spaces a horizontal tab should represent */
	int fixed_width;			/* TRUE will ignore any font variable width markers */
	int render;					/* FALSE will set next_char without rendering */
	char *text;					/* A pointer to the ASCII text to write */
	char *next_char;			/* ENG_WriteASCIIText sets this on exit */
} ENG_ASCIIText;

typedef struct engeffect {
	int effect_id;				/* An identifier that can be used to manipulate one or more effects */
	int klasse;					/* The type of effect */
	ENG_Object *object;			/* The primary object that the effect will er... affect */
	int frames;					/* The number of frames the effect will last for */
	int paused;					/* TRUE will pause this effect */
	int src_x;
	int src_y;
	int dst_x;
	int dst_y;
	int src_alpha;
	int src_alpha_over;
	int dst_alpha;
	int dst_alpha_over;
	/* Properties below are set by the system when the effect is performed */
	int current_frame;
} ENG_Effect;

/* Function prototypes */
/* NOTE TO SELF: As functions are placed here, check that the
   subsystem(s) they affect are operational */
int ENG_Init(ENG_OpEnv *op_env);												// [x] input validity checked
int ENG_Start(void);
int ENG_Quit(void);
char *ENG_GetError(void);
int ENG_ShowEvent(ENG_Event *event);											// [x] input validity checked

ENG_Object *ENG_GetObject(int id);												// [-][x] input validity/ENG_Start
ENG_Object *ENG_GetObjectUnderObject(ENG_Object *object, int hitpoint);			// [x][x] input validity/ENG_Start
int ENG_SetObjectZOrder(ENG_Object *object, int position, ENG_Object *target);	// [x][x] input validity/ENG_Start
int ENG_LoadObjects(int send_post_event);										// [-][x] input validity/ENG_Start
int ENG_UnloadObjects(char *objects_file);										// [-][x] input validity/ENG_Start
int ENG_SetObjectDefaults(ENG_Object *object);									// [x][-] input validity/ENG_Start
int ENG_CreateObject(ENG_Object *object, int ignore_image_dir);					// [x][x] input validity/ENG_Start
int ENG_FreeObject(ENG_Object *object, int free_images_only);					// [x][-] input validity/ENG_Start
int ENG_SendEvent(ENG_Event *event);											// [x][x] input validity/ENG_Start
int ENG_WriteASCIIText(ENG_ASCIIText *text, int page_line_or_word);				// [x][x] input validity/ENG_Start
int ENG_CreateEffect(ENG_Effect *effect);										// [x][x] input validity/ENG_Start
int ENG_ControlEffects(int effect_id, int state);								// [x][x] input validity/ENG_Start
int ENG_GetConfigOption(char *name, char *value);								// [x][x] input validity/ENG_Start
int ENG_SetConfigOption(char *name, char *value);								// [x][x] input validity/ENG_Start


/* TO DO */
int ENG_GetActionProperties(ENG_Action *action);
int ENG_SetActionProperties(ENG_Action *action);


int ENG_SetOpEnvImageDir(char *image_dir);
int ENG_SetOpEnvObjectsFile(char *objects_file);
int ENG_SetOpEnvConfigFile(char *config_file);
int ENG_SetOpEnvScreen(int screen_w, int screen_h);
int ENG_SetOpEnvCanvas(int canvas_w, int canvas_h);
int ENG_SetOpEnvViewport(int viewport_x, int viewport_y, int viewport_w, int viewport_h, int viewport_screen_x, int viewport_screen_y);
int ENG_SetOpEnvTitle(char *title);
int ENG_SetOpEnvFullScreen(int fullscreen);
int ENG_SetOpEnvSystemPointer(int systempointer);
int ENG_SetOpEnvPointer(int pointer, int pointer_mode);
int ENG_SetOpEnvPointerMoveUnits(int pointer_move_unit_low, int pointer_move_unit_high, int pointer_move_unit_threshold);
int ENG_SetOpEnvDelayAndCycleLimit(int delay, int cycle_limit);
int ENG_SetOpEnvEventProcessor(void (*event_processor)(ENG_Event *event));











