/* ENG is Not a Game (c) Thunor 2008 <thunorsif@hotmail.com>
www.thunor.org.uk

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. */


/* Includes */
#include "ENG_engine.h"


/* Defines */
/*
#define DEBUG_EFFECTS
*/

#define ENG_MAX_EFFECTS 1000


/* Variables */
ENG_Effect *eng_effects[ENG_MAX_EFFECTS];


/* Function prototypes */
int ENG_FreeEffect(ENG_Effect *effect);


/***************************************************************************
 * Initialise Effects                                                      *
 ***************************************************************************/

void ENG_InitEffects(void) {
	eng_effect_count = 0;
}

/***************************************************************************
 * Apply Effects                                                           *
 ***************************************************************************/
/* This applies effects to objects. If the effect is paused then it will
   continue once it has been unpaused.
   Effects are freed when their frames expire */
   
void ENG_ApplyEffects(void) {
	int count;
	ENG_Event event;

	for (count = eng_effect_count - 1; count >= 0; count--) {
		if (eng_effects[count]->paused == FALSE) {
			if (eng_effects[count]->klasse == ENG_EFFECT_LINEAR_MOVE) {
				if (eng_effects[count]->current_frame >= eng_effects[count]->frames) {
					/* Set the destination x/y values */
					eng_effects[count]->object->x = eng_effects[count]->dst_x;
					eng_effects[count]->object->y = eng_effects[count]->dst_y;
					/* The effect is complete so send an event and free it */		
					event.object = eng_effects[count]->object; event.klasse = ENG_EVENT_POST_EFFECT; ENG_SendEvent(&event);
					ENG_FreeEffect(eng_effects[count]);
				} else {
					/* Apply effect based upon frames processed */		
					eng_effects[count]->object->x = eng_effects[count]->src_x + (eng_effects[count]->dst_x - eng_effects[count]->src_x) / (float)eng_effects[count]->frames * eng_effects[count]->current_frame;
					eng_effects[count]->object->y = eng_effects[count]->src_y + (eng_effects[count]->dst_y - eng_effects[count]->src_y) / (float)eng_effects[count]->frames * eng_effects[count]->current_frame;
					/* Maintain a count of the frames processed */		
					eng_effects[count]->current_frame += 1;
				}
			} else if (eng_effects[count]->klasse == ENG_EFFECT_ALPHA_FADE) {
				if (eng_effects[count]->current_frame >= eng_effects[count]->frames) {
					/* Set the destination alpha/_over values */
					eng_effects[count]->object->alpha = eng_effects[count]->dst_alpha;
					eng_effects[count]->object->alpha_over = eng_effects[count]->dst_alpha_over;
					/* The effect is complete so send an event and free it */		
					event.object = eng_effects[count]->object; event.klasse = ENG_EVENT_POST_EFFECT; ENG_SendEvent(&event);
					ENG_FreeEffect(eng_effects[count]);
				} else {
					/* Apply effect based upon frames processed */		
					eng_effects[count]->object->alpha = eng_effects[count]->src_alpha + (eng_effects[count]->dst_alpha - eng_effects[count]->src_alpha) / (float)eng_effects[count]->frames * eng_effects[count]->current_frame;
					eng_effects[count]->object->alpha_over = eng_effects[count]->src_alpha_over + (eng_effects[count]->dst_alpha_over - eng_effects[count]->src_alpha_over) / (float)eng_effects[count]->frames * eng_effects[count]->current_frame;
					/* Maintain a count of the frames processed */		
					eng_effects[count]->current_frame += 1;
				}
			}
		}
	}
}

/***************************************************************************
 * Create Effect                                                           *
 ***************************************************************************/
/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */
	
int ENG_CreateEffect(ENG_Effect *effect) {
	int retval = ENG_ERROR_NONE;
	ENG_Effect *neweffect;

	if (eng_started == FALSE) {
		sprintf(eng_last_error, "%s: ENG has not been started yet", __func__);
		retval = ENG_ERROR_ENG_NOT_YET_STARTED;
	} else if (effect == NULL || sizeof(*effect) != sizeof(ENG_Effect)) {
		sprintf(eng_last_error, "%s: effect is invalid", __func__);
		retval = ENG_ERROR_INVALID_EFFECT;
	} else if (effect->object == NULL || sizeof(*effect->object) != sizeof(ENG_Object)) {
		sprintf(eng_last_error, "%s: effect->object is invalid", __func__);
		retval = ENG_ERROR_INVALID_OBJECT;
	} else if (eng_effect_count >= ENG_MAX_EFFECTS) {
		sprintf(eng_last_error, "%s: Too many effects (max %i)", __func__, ENG_MAX_EFFECTS);
		retval = ENG_ERROR_TOO_MANY_EFFECTS;
	} else {
		neweffect = (ENG_Effect*) malloc(sizeof(ENG_Effect));
		*neweffect = *effect;
		neweffect->current_frame = 0;
		#ifdef DEBUG_EFFECTS
			printf("%s: Creating %p effect_id=%i klasse=%i object->id=%i\n", __func__, neweffect, neweffect->effect_id, neweffect->klasse, neweffect->object->id);
		#endif
		eng_effects[eng_effect_count] = neweffect;
		eng_effect_count++;
	}

	return retval;
}

/***************************************************************************
 * Free Effect                                                             *
 ***************************************************************************/
/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */
	
int ENG_FreeEffect(ENG_Effect *effect) {
	int retval = ENG_ERROR_NONE, count, count2;

	if (eng_effect_count <= 0) {
		sprintf(eng_last_error, "%s: There is no effect to free", __func__);
		retval = ENG_ERROR_INVALID_EFFECT;
	} else if (effect == NULL || sizeof(*effect) != sizeof(ENG_Effect)) {
		sprintf(eng_last_error, "%s: effect is invalid", __func__);
		retval = ENG_ERROR_INVALID_EFFECT;
	} else {
		#ifdef DEBUG_EFFECTS
			printf("%s: Freeing %p effect_id=%i klasse=%i object->id=%i\n", __func__, effect, effect->effect_id, effect->klasse, effect->object->id);
		#endif
		/* If anything is above the current effect then move it down the array */
		for (count = 0; count < eng_effect_count - 1; count++) {
			if (eng_effects[count] == effect) {
				for (count2 = count; count2 < eng_effect_count - 1; count2++) {
					eng_effects[count2] = eng_effects[count2 + 1];
				}
				break;				
			}
		}
		/* Free the memory and update the array element count now */
		free(effect);
		eng_effect_count--;
	}

	return retval;
}

/***************************************************************************
 * Control Effects                                                         *
 ***************************************************************************/
/* This enables the user to apply any of several states to existing effects.
   
   On entry: effect_id = an identifier that represents one or more effects
			 state = ENG_EFFECT_STATE_PAUSE to pause one or more effects
			 state = ENG_EFFECT_STATE_UNPAUSE to resume one or more effects
			 state = ENG_EFFECT_STATE_FINALISE to take one or more effects
					 to their destination and then free them
			 state = ENG_EFFECT_STATE_TERMINATE to free one or more effects
	On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_ControlEffects(int effect_id, int state) {
	int retval = ENG_ERROR_NONE, count, found = FALSE;
	
	if (eng_started == FALSE) {
		sprintf(eng_last_error, "%s: ENG has not been started yet", __func__);
		retval = ENG_ERROR_ENG_NOT_YET_STARTED;
	} else if (state != ENG_EFFECT_STATE_PAUSE && state != ENG_EFFECT_STATE_UNPAUSE &&
			   state != ENG_EFFECT_STATE_FINALISE && state != ENG_EFFECT_STATE_TERMINATE) {
		sprintf(eng_last_error, "%s: state is invalid", __func__);
		retval = ENG_ERROR_INVALID_EFFECT_STATE;
	} else if (eng_effect_count <= 0) {
		sprintf(eng_last_error, "%s: There are no effects to control", __func__);
		retval = ENG_ERROR_INVALID_EFFECT;
	} else {
		/* Apply one of several states to one or more effects*/
		for (count = eng_effect_count - 1; count >= 0; count--) {
			if (eng_effects[count]->effect_id == effect_id) {
				found = TRUE;
				if (state == ENG_EFFECT_STATE_PAUSE) {
					eng_effects[count]->paused = TRUE;
				} else if (state == ENG_EFFECT_STATE_UNPAUSE) {
					eng_effects[count]->paused = FALSE;
				} else if (state == ENG_EFFECT_STATE_FINALISE) {
					eng_effects[count]->paused = FALSE;
					eng_effects[count]->current_frame = eng_effects[count]->frames;
				} else if (state == ENG_EFFECT_STATE_TERMINATE) {
					ENG_FreeEffect(eng_effects[count]);
				}
			}
		}
		if (found) {
			if (state == ENG_EFFECT_STATE_FINALISE) ENG_ApplyEffects();
		} else {
			sprintf(eng_last_error, "%s: effect_id is invalid", __func__);
			retval = ENG_ERROR_INVALID_EFFECT_ID;
		}
	}

	return retval;
}

/***************************************************************************
 * Free All Effects                                                        *
 ***************************************************************************/
/* This mops-up any remaining effects before shutting down */

void ENG_FreeAllEffects(void) {
	int count;
	
	if (eng_effect_count > 0) {
		printf("%s: Freeing\n", __func__);
		/* Free all remaining effects (ignoring errors) */
		for (count = eng_effect_count - 1; count >= 0; count--) {
			ENG_FreeEffect(eng_effects[count]);
		}
	}
}




