/* ENG is Not a Game (c) Thunor 2008 <thunorsif@hotmail.com>
www.thunor.org.uk

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. */


/* Includes */
#include "ENG_engine.h"


/* Defines */
/*
#define DEBUG_TIMED_EXIT
#define DEBUG_DUMP_VM_USAGE
*/
#define DEBUG_DUMP_VM_USAGE


/* Variables */
int eng_inited = FALSE;
int eng_last_cps = 0;


/* Function prototypes */
int ENG_MainLoop(void);
void ENG_Woden(void);
void ENG_CleanUp(void);


/***************************************************************************
 * Initialise                                                              *
 ***************************************************************************/
/* This is called once at the start. The user can load any config file
   after calling this function */

/* On entry: op_env points to the operating environment data */
/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_Init(ENG_OpEnv *op_env) {
	int retval = ENG_ERROR_NONE;

	if (eng_inited == FALSE) {
		printf("%s: Initialising\n", __func__);
		atexit(ENG_CleanUp);

		/* Validate operating environment data */
		retval = ENG_ValidateOpEnv(op_env, ENG_VALIDATE_ALL);
		if (!retval) {
			eng_op_env = *op_env;	/* Actually create a local copy of the passed op_env */

			/* Initialise some variables */
			eng_started = FALSE;
			eng_run = FALSE;
			eng_next_cycle = TRUE;
			
			/* Initialise all subsystems */
			ENG_InitOpEnv();
			ENG_InitConfig();
			ENG_InitSDL();		
			ENG_InitObjects();
			ENG_InitImages();
			ENG_InitActions();
			ENG_InitEvents();
			ENG_InitText();
			ENG_InitEffects();

			/* Initialise the random seed */
			srand(time(NULL));

			/* Load the config file */
			if (strcmp(eng_op_env.config_file, "") != 0) retval = ENG_LoadConfig();

			eng_inited = TRUE;
		}
	} else {
		sprintf(eng_last_error, "%s: ENG already initialised", __func__);
		retval = ENG_ERROR_ENG_ALREADY_INITIALISED;
	}
	
	return retval;
}

/***************************************************************************
 * Start                                                                   *
 ***************************************************************************/
/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_Start(void) {
	int retval = ENG_ERROR_NONE;
	
	if (eng_inited == TRUE && eng_started == FALSE) {
		eng_started = TRUE;
		printf("%s: Starting\n", __func__);

		retval = ENG_StartSDL();
		if (!retval) {
			retval = ENG_LoadObjects(TRUE);
			if (!retval) {
				retval = ENG_MainLoop();
				ENG_FreeAllEffects();
			}
			ENG_FreeAllObjects();
			ENG_FreeAllImages();
		}
		ENG_StopSDL();
		/* Save the config file */
		if (strcmp(eng_op_env.config_file, "") != 0) retval = ENG_SaveConfig();
		ENG_FreeAllConfigOptions();

		eng_started = FALSE;
		eng_inited = FALSE;
	} else if (eng_inited == FALSE) {
		sprintf(eng_last_error, "%s: ENG has not been initialised yet", __func__);
		retval = ENG_ERROR_ENG_NOT_YET_INITIALISED;
	} else {
		sprintf(eng_last_error, "%s: ENG already started", __func__);
		retval = ENG_ERROR_ENG_ALREADY_STARTED;
	}

	return retval;
}

/***************************************************************************
 * Main Loop                                                               *
 ***************************************************************************/
/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_MainLoop(void) {
	int retval = ENG_ERROR_NONE;
	#ifdef DEBUG_DUMP_VM_USAGE
		char vmusage[80];
	#endif

	printf("%s: Looping...\n", __func__);

	eng_run = TRUE;
	while (eng_run && !retval) {
		ENG_GenerateEvents();
		ENG_ApplyEffects();
		ENG_Woden();
		retval = ENG_RenderObjects();
		if (!retval) retval = ENG_UpdateScreen(eng_screen);
	}
	#ifdef DEBUG_DUMP_VM_USAGE
		/*sprintf(vmusage, "cat /proc/%u/status | grep VmSize", getpid());*/
		sprintf(vmusage, "cat /proc/%u/status | grep Vm > vmusage.txt", getpid());
		system (vmusage);
	#endif

	return retval;
}

/***************************************************************************
 * Woden                                                                   *
 ***************************************************************************/
/* Woden the Ræfntíw, assisted by Hugin and Munin manage the engine */

void ENG_Woden(void) {
	static int init = TRUE, startticks, lastticks, cycles = 0;
	static int last_cps = 0;
	static ENG_ASCIIText ASCIIText_cps;
	static char cps_text[8];
	
	/* Intialise some variables the once */
	if (init) {
		init = FALSE;
		/* Initialise the cps timer variables */
		startticks = lastticks = SDL_GetTicks();
		/* Initialise ASCIIText_cps.
		   Note that the font will be loaded else ENG would have quit at ENG_Start */
		ASCIIText_cps.font = ENG_GetObject(ENG_FNT_ASCII_8X12_WHITE_BLACK_OUTLINE)->img_surface;
		ASCIIText_cps.target = eng_lyr_debug->img_surface;
		ASCIIText_cps.img_x = eng_lyr_debug->img_x;
		ASCIIText_cps.img_y = eng_lyr_debug->img_y;
		ASCIIText_cps.img_w = eng_lyr_debug->img_w;
		ASCIIText_cps.img_h = eng_lyr_debug->img_h;
		ASCIIText_cps.bgcolour = (ENG_COLOUR_KEY_RED << 16) | (ENG_COLOUR_KEY_GREEN << 8) | (ENG_COLOUR_KEY_BLUE);
		ASCIIText_cps.char_spacing = -1;	/* The default for the above sized font */
		ASCIIText_cps.line_spacing = 0;
		ASCIIText_cps.tab_spaces = 8;
		ASCIIText_cps.fixed_width = FALSE;
		ASCIIText_cps.render = TRUE;
		ASCIIText_cps.text = cps_text;
	}

	/* Is there a cycle limit in place? */
	if (eng_op_env.cycle_limit) {
		/* Wait until the timer sets the flag before proceeding */
		while (!eng_next_cycle) SDL_Delay(0);
		eng_next_cycle = FALSE;
	} else {
		/* If the user wants to delay every cycle then do so */
		if (eng_op_env.delay) SDL_Delay(eng_op_env.delay);
	}

	/* Count the cycles and record the cps */
	cycles++;
	if (SDL_GetTicks() - lastticks >= 1000) {
		lastticks = SDL_GetTicks();
		eng_last_cps = cycles;
		cycles = 0;
	}

	/* Update the cps text if different */
	if (last_cps != eng_last_cps && ENG_GetVisibility(eng_lyr_debug)) {
		last_cps = eng_last_cps;
		sprintf(cps_text, "%i", last_cps);
		ENG_WriteASCIIText(&ASCIIText_cps, ENG_TEXT_WRITE_PAGE);
	}

	#ifdef DEBUG_TIMED_EXIT
		if (SDL_GetTicks() - startticks >= 20000) ENG_Quit();
	#endif

	/* Has the main loop been stopped? */
	if (!eng_run) {
	}
	
}

/***************************************************************************
 * Quit (ENG)                                                              *
 ***************************************************************************/
/* This is how it works :-

   If the user has only called ENG_Init then ENG_FreeAllConfigOptions unloads
   the only subsystem to have been initialised at this point.
   
   If the user has also called ENG_Start then firstly an ENG_EVENT_QUITTING
   event is sent to the user's event processor and then all subsystems are
   unloaded when returning to and dropping out of ENG_Start */

/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_Quit(void) {
	int retval = ENG_ERROR_NONE;
	ENG_Event event;

	if (eng_started == TRUE) {
		printf("%s: Quitting\n", __func__);
		event.object = eng_lyr_root; event.klasse = ENG_EVENT_QUITTING; ENG_SendEvent(&event);
		eng_run = FALSE;
	} else if (eng_inited == TRUE) {
		printf("%s: Quitting\n", __func__);
		/* Save the config file */
		if (strcmp(eng_op_env.config_file, "") != 0) retval = ENG_SaveConfig();
		ENG_FreeAllConfigOptions();
		eng_inited = FALSE;
	} else {
		sprintf(eng_last_error, "%s: There is nothing to quit: ENG has not been initialised yet", __func__);
		retval = ENG_ERROR_ENG_NOT_YET_INITIALISED;
	}

	return retval;
}

/***************************************************************************
 * Get Error                                                               *
 ***************************************************************************/
/*  On exit: returns a pointer to a string containing the last internal
			 error plus the contents of SDL_GetError if relevant */

char *ENG_GetError(void) {
	static char last_error[ENG_ERROR_LENGTH];
	
	strcpy (last_error, eng_last_error);
	strcpy (eng_last_error, "");
	
	return last_error;
}

/***************************************************************************
 * Clean Up                                                                *
 ***************************************************************************/
/* This function is registered with atexit and is a fail safe.
   All the functions listed can be called multiple times without problem */

void ENG_CleanUp(void) {
	printf("%s: Cleaning-up\n", __func__);
	ENG_FreeAllEffects();
	ENG_FreeAllObjects();
	ENG_FreeAllImages();
	ENG_StopSDL();
	ENG_FreeAllConfigOptions();
}



