/* ENG is Not a Game (c) Thunor 2008 <thunorsif@hotmail.com>
www.thunor.org.uk

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. */


/* Includes */
#include "ENG_engine.h"


/* Defines */
/*
#define DEBUG_OBJECTS
#define DEBUG_OBJECTS_ALL_DRAGGABLE
#define DEBUG_OBJECTS_FILE
#define DEBUG_OBJECT_SORT
#define DEBUG_OBJECT_ZORDER
#define DEBUG_GET_ENABILITY
#define DEBUG_GET_VISIBILITY
#define DEBUG_FIND_ANCESTOR
*/

#define ENG_MAX_OBJECTS 1000

#define ENG_REC_POINTER_X 4
#define ENG_REC_POINTER_Y 4
#define ENG_REC_POINTER_ALPHA 50


/* Variables */
ENG_Object *eng_objects[ENG_MAX_OBJECTS];
ENG_Object *eng_objects_id[ENG_MAX_OBJECTS];


/* Function prototypes */
int ENG_FillRectA(SDL_Surface *surface, SDL_Rect *dstrect, Uint32 colour, Uint8 alpha);


/***************************************************************************
 * Initialise Objects                                                      *
 ***************************************************************************/

void ENG_InitObjects(void) {
	eng_object_count = 0;
	eng_oup = NULL;
	eng_pressed_object = NULL;
}

/***************************************************************************
 * Render Objects                                                          *
 ***************************************************************************/
/* Note: No clearing of the screen or viewport is done. If the screen or
   viewport requires setting to a single colour then this can be achieved by
   setting ENG_LYR_SCREEN's or ENG_LYR_VIEWPORT's bgcolour/bgcolour_over
   properties and the renderer will fill these areas */
   
/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */
	
int ENG_RenderObjects(void) {
	int count, retval = ENG_ERROR_NONE;
	int cx, cy, overeventfound, ix, iy, iw, ih, alpha, alpha_over, finalalpha;
	SDL_Rect srcrect, dstrect;
	SDL_Surface *img_surface;
	Uint32 bgcolour, use_rleaccel = 0;
	#ifdef ENG_EVENT_SEND_PRE_RENDER
		ENG_Event event;

		/* The user may want to do something beforehand */
		event.object = eng_lyr_root; event.klasse = ENG_EVENT_PRE_RENDER; ENG_SendEvent(&event);
	#endif
	
	/* Iterate through the objects */
	for (count = 0; count < eng_object_count; count++) {
		/* Get the object's visibility */
		if (ENG_GetVisibility(eng_objects[count])) {

			/* Is there an over event waiting for this object? */
			if (eng_objects[count] == eng_oup && ENG_GetEnability(eng_objects[count])) {
				overeventfound = TRUE;
			} else {
				overeventfound = FALSE;
			}

			/* Set the screen and blend screen clipping rectangles */
			/* Trace back through the object's [p]ids searching for ENG_LYR_SCREEN */
			if (ENG_FindAncestor(eng_objects[count], eng_lyr_screen)) {
				/* Set clipping rectangles to screen */
				dstrect.x = 0;
				dstrect.y = 0;
				dstrect.w = eng_op_env.screen_w;
				dstrect.h = eng_op_env.screen_h;
			} else {
				/* Set clipping rectangles to viewport */
				dstrect.x = eng_op_env.viewport_screen_x;
				dstrect.y = eng_op_env.viewport_screen_y;
				dstrect.w = eng_op_env.viewport_w;
				dstrect.h = eng_op_env.viewport_h;
			}
			SDL_SetClipRect(eng_screen, &dstrect);
			SDL_SetClipRect(eng_blend_screen, &dstrect);

			/* Get object's canvas offset */
			ENG_GetCanvasOffset(eng_objects[count], &cx, &cy);

			/* Get object's alpha and alpha_over */
			alpha = ENG_GetAlpha(eng_objects[count]);
			alpha_over = ENG_GetAlphaOver(eng_objects[count]);

			/** bgcolour and bgcolour_over */
			if ((!overeventfound && eng_objects[count]->bgcolour != ENG_UNDEFINED) || 
				(overeventfound && eng_objects[count]->bgcolour_over != ENG_UNDEFINED)) {
				if ((!overeventfound && alpha == 255) || (overeventfound && alpha_over == 255)) {
					if (!overeventfound && eng_objects[count]->bgcolour != ENG_UNDEFINED) {
						bgcolour = SDL_MapRGB(eng_screen->format, (eng_objects[count]->bgcolour >> 16) & 0xff, (eng_objects[count]->bgcolour >> 8) & 0xff, eng_objects[count]->bgcolour & 0xff);
					} else {
						bgcolour = SDL_MapRGB(eng_screen->format, (eng_objects[count]->bgcolour_over >> 16) & 0xff, (eng_objects[count]->bgcolour_over >> 8) & 0xff, eng_objects[count]->bgcolour_over & 0xff);
					}
					/* Fill rectangle */
					dstrect.x = cx - eng_op_env.viewport_x + eng_op_env.viewport_screen_x;
					dstrect.y = cy - eng_op_env.viewport_y + eng_op_env.viewport_screen_y;
					dstrect.w = eng_objects[count]->w; dstrect.h = eng_objects[count]->h;
					if ((SDL_FillRect(eng_screen, &dstrect, bgcolour)) < 0) {
						sprintf(eng_last_error, "%s: Cannot fill surface: %s", __func__, SDL_GetError());
						retval = ENG_ERROR_CANNOT_FILL_SURFACE;
						break;
					}
				} else {
					if (!overeventfound && eng_objects[count]->bgcolour != ENG_UNDEFINED) {
						bgcolour = SDL_MapRGB(eng_blend_screen->format, (eng_objects[count]->bgcolour >> 16) & 0xff, (eng_objects[count]->bgcolour >> 8) & 0xff, eng_objects[count]->bgcolour & 0xff);
						finalalpha = alpha;
					} else {
						bgcolour = SDL_MapRGB(eng_blend_screen->format, (eng_objects[count]->bgcolour_over >> 16) & 0xff, (eng_objects[count]->bgcolour_over >> 8) & 0xff, eng_objects[count]->bgcolour_over & 0xff);
						finalalpha = alpha_over;
					}
					/* Fill rectangle */
					dstrect.x = cx - eng_op_env.viewport_x + eng_op_env.viewport_screen_x;
					dstrect.y = cy - eng_op_env.viewport_y + eng_op_env.viewport_screen_y;
					dstrect.w = eng_objects[count]->w; dstrect.h = eng_objects[count]->h;
					retval = ENG_FillRectA(eng_screen, &dstrect, bgcolour, finalalpha);
					if (retval) break;
				}
			}

			/** img_file and img_over_file */
			if ((!overeventfound && eng_objects[count]->img_surface != NULL) ||
				(overeventfound && eng_objects[count]->img_over_surface != NULL)) {
				if (!overeventfound && eng_objects[count]->img_surface != NULL) {
					ix = eng_objects[count]->img_x; iy = eng_objects[count]->img_y;
					iw = eng_objects[count]->img_w; ih = eng_objects[count]->img_h;
					img_surface = eng_objects[count]->img_surface;
					finalalpha = alpha;
				} else {
					ix = eng_objects[count]->img_over_x; iy = eng_objects[count]->img_over_y;
					iw = eng_objects[count]->img_over_w; ih = eng_objects[count]->img_over_h;
					img_surface = eng_objects[count]->img_over_surface;
					finalalpha = alpha_over;
				}
				/* If image region is unused then use entire image */
				if (ix == ENG_UNDEFINED && iy == ENG_UNDEFINED && iw == ENG_UNDEFINED && ih == ENG_UNDEFINED) {
					ix = 0; iy = 0;
					iw = img_surface->w; ih = img_surface->h;
				}
				if (finalalpha != 255) {
					/* Record surface's use of SDL_RLEACCEL.
					   Because the surface is not compressed until the first blit,
					   SDL_RLEACCELOK must be checked for too, I imagine as compression is pending */
					if (img_surface->flags & (SDL_RLEACCEL | SDL_RLEACCELOK)) use_rleaccel = SDL_RLEACCEL;
					/* Set the image surface alpha. */
					/* Note that this removes SDL_RLEACCEL because using it with alpha
					   is very very slow; it's restored at the end of the function */
					if ((SDL_SetAlpha(img_surface, SDL_SRCALPHA, finalalpha)) < 0) {
						sprintf(eng_last_error, "%s: Cannot set surface alpha: %s", __func__, SDL_GetError());
						retval = ENG_ERROR_CANNOT_SET_SURFACE_ALPHA;
						break;
					}
				}
				srcrect.x = ix; srcrect.y = iy; srcrect.w = iw; srcrect.h = ih;
				dstrect.x = cx - eng_op_env.viewport_x + eng_op_env.viewport_screen_x;
				dstrect.y = cy - eng_op_env.viewport_y + eng_op_env.viewport_screen_y;
				dstrect.w = eng_objects[count]->w; dstrect.h = eng_objects[count]->h;
				/* Blit surface */
				if ((SDL_BlitSurface(img_surface, &srcrect, eng_screen, &dstrect)) < 0) {
					sprintf(eng_last_error, "%s: Cannot blit surface: %s", __func__, SDL_GetError());
					retval = ENG_ERROR_CANNOT_BLIT_SURFACE;
					break;
				}
				if (finalalpha != 255) {
					/* Disable the image surface alpha and restore SDL_RLEACCEL if set.
					   Note that officially NOT passing the SDL_SRCALPHA flag disables alpha blits from this
					   surface but then it's not possible to reenable SDL_RLEACCEL, so I simply set the alpha
					   value to SDL_ALPHA_OPAQUE (255) which achieves the same result. */
					if ((SDL_SetAlpha(img_surface, SDL_SRCALPHA | use_rleaccel, 255)) < 0) {
						sprintf(eng_last_error, "%s: Cannot set surface alpha: %s", __func__, SDL_GetError());
						retval = ENG_ERROR_CANNOT_SET_SURFACE_ALPHA;
						break;
					}
				}
			}
		}
	}
		
	/* Reset screen and blend screen clipping rectangles */
	SDL_SetClipRect(eng_screen, NULL);
	SDL_SetClipRect(eng_blend_screen, NULL);

	return retval;
}

/***************************************************************************
 * FillRectA                                                               *
 ***************************************************************************/
/* SDL_BlitSurface alpha blends, so an alpha rectangle first needs to become
   a surface with per-surface-alpha set and then blitted to the screen */

/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_FillRectA(SDL_Surface *surface, SDL_Rect *dstrect, Uint32 colour, Uint8 alpha) {
	int retval = ENG_ERROR_NONE;
	
	/* Set the blend surface alpha */
	if ((SDL_SetAlpha(eng_blend_screen, SDL_SRCALPHA, alpha)) < 0) {
		sprintf(eng_last_error, "%s: Cannot set surface alpha: %s", __func__, SDL_GetError());
		retval = ENG_ERROR_CANNOT_SET_SURFACE_ALPHA;
	}
	if (!retval) {
		/* Fill rectangle */
		if ((SDL_FillRect(eng_blend_screen, dstrect, colour)) < 0) {
			sprintf(eng_last_error, "%s: Cannot fill surface: %s", __func__, SDL_GetError());
			retval = ENG_ERROR_CANNOT_FILL_SURFACE;
		}
	}
	if (!retval) {
		/* Blit surface */
		if ((SDL_BlitSurface(eng_blend_screen, dstrect, surface, dstrect)) < 0) {
			sprintf(eng_last_error, "%s: Cannot blit surface: %s", __func__, SDL_GetError());
			retval = ENG_ERROR_CANNOT_BLIT_SURFACE;
		}
	}

	return retval;
}

/***************************************************************************
 * Update Screen                                                           *
 ***************************************************************************/
/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_UpdateScreen(SDL_Surface *surface) {
	int retval = ENG_ERROR_NONE;
	#ifdef ENG_EVENT_SEND_POST_RENDER
		ENG_Event event;

		/* The user may want to do something beforehand */
		event.object = eng_lyr_root; event.klasse = ENG_EVENT_POST_RENDER; ENG_SendEvent(&event);
	#endif
	
	if ((SDL_Flip(surface)) == -1) {
		sprintf(eng_last_error, "%s: Cannot update screen surface: %s", __func__, SDL_GetError());
		retval = ENG_ERROR_CANNOT_UPDATE_SURFACE;
	}
	
	return retval;
}

/***************************************************************************
 * Get Canvas Offset                                                       *
 ***************************************************************************/
/* This gets an object's canvas offset.
   If a NULL pob is found then it will be interpreted as eng_lyr_canvas */

int ENG_GetCanvasOffset(ENG_Object *object, int *cx, int *cy) {
	int retval = ENG_ERROR_NONE;

	if (object == NULL || sizeof(*object) != sizeof(ENG_Object)) {
		sprintf(eng_last_error, "%s: object is invalid", __func__);
		retval = ENG_ERROR_INVALID_OBJECT;
	} else {
		*cx = object->x;
		*cy = object->y;
		while (object != eng_lyr_canvas && object != eng_lyr_root) {
			if (object->pob == NULL) {
				object = eng_lyr_canvas;
			} else {
				object = object->pob;
			}
			*cx += object->x;
			*cy += object->y;
		}
	}
	
	return retval;
}

/***************************************************************************
 * Get Visibility                                                          *
 ***************************************************************************/
/* This gets an object's visibility.
   If a NULL pob is found then it will be interpreted as eng_lyr_canvas.
   Note this could return ENG_INHERITED if the user has set the canvas or
   root objects to ENG_INHERITED but this is unlikely ;) */
   
/*  On exit: returns TRUE or FALSE */

int ENG_GetVisibility(ENG_Object *object) {
	
	if (object == NULL || sizeof(*object) != sizeof(ENG_Object)) {
		sprintf(eng_last_error, "%s: object is invalid", __func__);
	} else {
		#ifdef DEBUG_GET_VISIBILITY
			printf("\n%s: Visibility for object %i\n", __func__, object->id);
			printf("%s: id=%i pob=%p visible=%i\n", __func__, object->id, object->pob, object->visible);
		#endif
		while (object->visible == ENG_INHERITED) {
			if (object == eng_lyr_root || object == eng_lyr_canvas) {
				break;
			} else {
				if (object->pob == NULL) {
					object = eng_lyr_canvas;
				} else {
					object = object->pob;
				}
			}
			#ifdef DEBUG_GET_VISIBILITY
				printf("%s: id=%i pob=%p visible=%i\n", __func__, object->id, object->pob, object->visible);
			#endif
		}
	}

	return object->visible;
}

/***************************************************************************
 * Get Enability                                                           *
 ***************************************************************************/
/* This gets an object's enability.
   If a NULL pob is found then it will be interpreted as eng_lyr_canvas.
   Note this could return ENG_INHERITED if the user has set the canvas or
   root objects to ENG_INHERITED but this is unlikely ;) */
   
/*  On exit: returns TRUE or FALSE */

int ENG_GetEnability(ENG_Object *object) {
	
	if (object == NULL || sizeof(*object) != sizeof(ENG_Object)) {
		sprintf(eng_last_error, "%s: object is invalid", __func__);
	} else {
		#ifdef DEBUG_GET_ENABILITY
			printf("\n%s: Enability for object %i\n", __func__, object->id);
			printf("%s: id=%i pob=%p enabled=%i\n", __func__, object->id, object->pob, object->enabled);
		#endif
		while (object->enabled == ENG_INHERITED) {
			if (object == eng_lyr_root || object == eng_lyr_canvas) {
				break;
			} else {
				if (object->pob == NULL) {
					object = eng_lyr_canvas;
				} else {
					object = object->pob;
				}
			}
			#ifdef DEBUG_GET_ENABILITY
				printf("%s: id=%i pob=%p enabled=%i\n", __func__, object->id, object->pob, object->enabled);
			#endif
		}
	}

	return object->enabled;
}

/***************************************************************************
 * Get Alpha                                                               *
 ***************************************************************************/
/* This gets an object's alpha.
   If a NULL pob is found then it will be interpreted as eng_lyr_canvas.
   Note this could return ENG_INHERITED if the user has set the canvas or
   root objects to ENG_INHERITED but this is unlikely ;) */
   
/*  On exit: returns 0 to 255 */

int ENG_GetAlpha(ENG_Object *object) {
	
	if (object == NULL || sizeof(*object) != sizeof(ENG_Object)) {
		sprintf(eng_last_error, "%s: object is invalid", __func__);
	} else {
		while (object->alpha == ENG_INHERITED) {
			if (object == eng_lyr_root || object == eng_lyr_canvas) {
				break;
			} else {
				if (object->pob == NULL) {
					object = eng_lyr_canvas;
				} else {
					object = object->pob;
				}
			}
		}
	}

	return object->alpha;
}

/***************************************************************************
 * Get Alpha Over                                                          *
 ***************************************************************************/
/* This gets an object's alpha_over.
   If a NULL pob is found then it will be interpreted as eng_lyr_canvas.
   Note this could return ENG_INHERITED if the user has set the canvas or
   root objects to ENG_INHERITED but this is unlikely ;) */
   
/*  On exit: returns 0 to 255 */

int ENG_GetAlphaOver(ENG_Object *object) {
	
	if (object == NULL || sizeof(*object) != sizeof(ENG_Object)) {
		sprintf(eng_last_error, "%s: object is invalid", __func__);
	} else {
		while (object->alpha_over == ENG_INHERITED) {
			if (object == eng_lyr_root || object == eng_lyr_canvas) {
				break;
			} else {
				if (object->pob == NULL) {
					object = eng_lyr_canvas;
				} else {
					object = object->pob;
				}
			}
		}
	}

	return object->alpha_over;
}

/***************************************************************************
 * Find Ancestor                                                           *
 ***************************************************************************/
/* This traces back from (and including) an object looking for a
   particular ancestor.
   If a NULL pob is found then it will be interpreted as eng_lyr_canvas */

/*  On exit: returns TRUE if found else FALSE */

int ENG_FindAncestor(ENG_Object *object, ENG_Object *target) {

	if (object == NULL || sizeof(*object) != sizeof(ENG_Object)) {
		sprintf(eng_last_error, "%s: object is invalid", __func__);
	} else if (target == NULL || sizeof(*target) != sizeof(ENG_Object)) {
		sprintf(eng_last_error, "%s: target is invalid", __func__);
	} else {
		#ifdef DEBUG_FIND_ANCESTOR
			printf("\n%s: Looking for object %i's ancestor %i\n", __func__, object->id, target->id);
			printf("%s: id=%i pob=%p\n", __func__, object->id, object->pob);
		#endif
		while (object != target) {
			if (object == eng_lyr_root || object == eng_lyr_canvas) {
				return FALSE;
			} else {
				if (object->pob == NULL) {
					object = eng_lyr_canvas;
				} else {
					object = object->pob;
				}
			}
			#ifdef DEBUG_FIND_ANCESTOR
				printf("%s: id=%i pob=%p\n", __func__, object->id, object->pob);
			#endif
		}
	}

	return TRUE;
}

/***************************************************************************
 * Get Object Under Object                                                 *
 ***************************************************************************/
/*  On exit: returns a pointer to the visible object under object on success else NULL */

ENG_Object *ENG_GetObjectUnderObject(ENG_Object *object, int hitpoint) {
	int count, cx, cy, cx2, cy2;
	
	if (eng_started == FALSE) {
		sprintf(eng_last_error, "%s: ENG has not been started yet", __func__);
	} else if (object == NULL || sizeof(*object) != sizeof(ENG_Object)) {
		sprintf(eng_last_error, "%s: object is invalid", __func__);
	} else if (hitpoint != ENG_HIT_POINT_TOP_LEFT && hitpoint != ENG_HIT_POINT_CENTRE) {
		sprintf(eng_last_error, "%s: hitpoint is invalid", __func__);
	} else {
		/* Get hit point of source object */
		ENG_GetCanvasOffset(object, &cx, &cy);
		if (hitpoint == ENG_HIT_POINT_TOP_LEFT) {
			/* cx and cy are ok */
		} else if (hitpoint == ENG_HIT_POINT_CENTRE) {
			cx += object->w / 2;		
			cy += object->h / 2;		
		}
		for (count = eng_object_count - 1; count >= 0; count--) {
			if (ENG_GetVisibility(eng_objects[count]) && count < object->zorder) {
				ENG_GetCanvasOffset(eng_objects[count], &cx2, &cy2);
				if (cx >= cx2 && cx < cx2 + eng_objects[count]->w && cy >= cy2 && cy < cy2 + eng_objects[count]->h) {
					return eng_objects[count];
				}
			}
		}
	}

	return NULL;
}

/***************************************************************************
 * Set Object ZOrder                                                       *
 ***************************************************************************/
/* User objects must be above ENG_LYR_VIEWPORT and below ENG_REC_POINTER to
   be accessible via the pointer. This function does not place any restrictions
   on the destinaton of any objects.

   Positions :-

   ENG_ZORDER_ABOVE_ALL
   --------------------
   The object will be moved above the target and above all other objects who's
   pob is also the target.
   If the object is or results in being the target then it is not moved.

   ENG_ZORDER_ABOVE
   ----------------
   The object will be moved above the target.
   If the object is the target then it is moved above itself.
   
   ENG_ZORDER_BELOW
   ----------------
   The object will be moved below the target.
   If the object is the target then it is moved below itself */
   
/* On entry: object is the object to move
			 position is one of the zorder positions detailed above
			 target is the target object of the move
    On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_SetObjectZOrder(ENG_Object *object, int position, ENG_Object *target) {
	int count, retval = ENG_ERROR_NONE;
	int index = object->zorder, targetindex = ENG_UNDEFINED, direction = 0;
	ENG_Object *swap_object;

	if (eng_started == FALSE) {
		sprintf(eng_last_error, "%s: ENG has not been started yet", __func__);
		retval = ENG_ERROR_ENG_NOT_YET_STARTED;
	} else if (object == NULL || sizeof(*object) != sizeof(ENG_Object)) {
		sprintf(eng_last_error, "%s: object is invalid", __func__);
		retval = ENG_ERROR_INVALID_OBJECT;
	} else if (target == NULL || sizeof(*target) != sizeof(ENG_Object)) {
		sprintf(eng_last_error, "%s: target is invalid", __func__);
		retval = ENG_ERROR_INVALID_TARGET_OBJECT;
	} else if (position != ENG_ZORDER_ABOVE_ALL && position != ENG_ZORDER_ABOVE && position != ENG_ZORDER_BELOW) {
		sprintf(eng_last_error, "%s: zorder position is invalid", __func__);
		retval = ENG_ERROR_INVALID_ZORDER_POSITION;
	} else {
		/* Set the target index */
		if (position == ENG_ZORDER_ABOVE_ALL) {
			/* Find the topmost object equal to target or its pob is equal to target */
			for (count = 0; count < eng_object_count; count++) {
				if (eng_objects[count] == target || eng_objects[count]->pob == target) targetindex = count;
			}
			/* Make some positional adjustments to the target index */
			if (targetindex == ENG_UNDEFINED) {
				targetindex = object->zorder;	/* No seg faults please :) */
			} else if (object->zorder > targetindex) {
				targetindex++;
			}
		} else if (position == ENG_ZORDER_ABOVE) {
			/* The target index is the zorder within the target */
			targetindex = target->zorder;
			/* Make some positional adjustments to the target index */
			if (object->zorder == targetindex && targetindex < eng_object_count - 1) {
				targetindex++;
			} else if (object->zorder > targetindex) {
				targetindex++;
			}
		} else if (position == ENG_ZORDER_BELOW) {
			/* The target index is the zorder within the target */
			targetindex = target->zorder;
			/* Make some positional adjustments to the target index */
			if (object->zorder == targetindex && targetindex > 0) {
				targetindex--;
			} else if (object->zorder < targetindex) {
				targetindex--;
			}
		}
		/* Set move direction */
		if (index < targetindex) {
			direction = 1;
		} else if (object->zorder > targetindex) {
			direction = -1;
		}
		#ifdef DEBUG_OBJECT_ZORDER
			printf("%s: Moving id=%i zorder=%i targetindex=%i (eng_object_count=%i)\n", __func__, object->id, object->zorder, targetindex, eng_object_count);
		#endif
		while (index != targetindex) {
			#ifdef DEBUG_OBJECT_ZORDER
				printf("    id=%i pob=%p zorder=%i <--> id=%i pob=%p zorder=%i\n", eng_objects[index]->id, eng_objects[index]->pob, eng_objects[index]->zorder, eng_objects[index + direction]->id, eng_objects[index + direction]->pob, eng_objects[index + direction]->zorder);
			#endif
			swap_object = eng_objects[index];
			eng_objects[index] = eng_objects[index + direction];
			eng_objects[index]->zorder -= direction;
			index += direction;
			eng_objects[index] = swap_object;
			eng_objects[index]->zorder += direction;
		}
	}

	return retval;
}

/***************************************************************************
 * Load Objects                                                            *
 ***************************************************************************/
/* The first time this is run it will create some system objects */
   
/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_LoadObjects(int send_post_event) {
	static int init = TRUE;
	int setdefaults = TRUE, retval = ENG_ERROR_NONE, quit = FALSE;
	int count, linecount = 0, objectline = 0, loadcount = 0;
	char line[256];
	char last_error[ENG_ERROR_LENGTH];
	ENG_Object object;
	ENG_Event event;
	FILE *fp;
	
	if (eng_started == FALSE) {
		sprintf(eng_last_error, "%s: ENG has not been started yet", __func__);
		retval = ENG_ERROR_ENG_NOT_YET_STARTED;
	} else {
		#ifdef DEBUG_OBJECTS
			printf("%s: Loading\n", __func__);
		#endif
		
		if (init) {
			/* First time in so create the system objects */
			init = FALSE;
			/* Create a root object */
			ENG_SetObjectDefaults(&object);
			object.id = ENG_LYR_ROOT;
			object.pid = ENG_LYR_ROOT;
			object.klasse = ENG_OBJECT_KLASSE_LAYER;
			object.visible = TRUE;
			object.enabled = TRUE;
			object.alpha = 255;
			object.alpha_over = 255;
			object.navigable = FALSE;
			object.source = ENG_OBJECT_SOURCE_SYSTEM;
			retval = ENG_CreateObject(&object, FALSE);
			if (!retval) {
				/* Create a canvas object */
				ENG_SetObjectDefaults(&object);
				object.id = ENG_LYR_CANVAS;
				object.pid = ENG_LYR_ROOT;
				object.klasse = ENG_OBJECT_KLASSE_LAYER;
				object.visible = TRUE;
				object.enabled = TRUE;
				object.alpha = 255;
				object.alpha_over = 255;
				object.w = eng_op_env.canvas_w;
				object.h = eng_op_env.canvas_h;
				object.navigable = FALSE;
				object.source = ENG_OBJECT_SOURCE_SYSTEM;
				retval = ENG_CreateObject(&object, FALSE);
			}
			if (!retval) {
				/* Create a screen object */
				ENG_SetObjectDefaults(&object);
				object.id = ENG_LYR_SCREEN;
				object.pid = ENG_LYR_CANVAS;
				object.klasse = ENG_OBJECT_KLASSE_LAYER;
				object.visible = TRUE;
				object.enabled = TRUE;
				object.alpha = 255;
				object.alpha_over = 255;
				object.x = eng_op_env.viewport_x - eng_op_env.viewport_screen_x;
				object.y = eng_op_env.viewport_y - eng_op_env.viewport_screen_y;
				object.w = eng_op_env.screen_w;
				object.h = eng_op_env.screen_h;
				object.navigable = FALSE;
				object.source = ENG_OBJECT_SOURCE_SYSTEM;
				retval = ENG_CreateObject(&object, FALSE);
			}
			if (!retval) {
				/* Create a viewport object */
				ENG_SetObjectDefaults(&object);
				object.id = ENG_LYR_VIEWPORT;
				object.pid = ENG_LYR_SCREEN;
				object.klasse = ENG_OBJECT_KLASSE_LAYER;
				object.visible = TRUE;
				object.enabled = TRUE;
				object.alpha = 255;
				object.alpha_over = 255;
				object.x = eng_op_env.viewport_screen_x;
				object.y = eng_op_env.viewport_screen_y;
				object.w = eng_op_env.viewport_w;
				object.h = eng_op_env.viewport_h;
				object.navigable = FALSE;
				object.source = ENG_OBJECT_SOURCE_SYSTEM;
				retval = ENG_CreateObject(&object, FALSE);
			}
			if (!retval) {
				/* Create a debug object */
				ENG_SetObjectDefaults(&object);
				object.id = ENG_LYR_DEBUG;
				object.pid = ENG_LYR_SCREEN;
				object.klasse = ENG_OBJECT_KLASSE_LAYER;
				object.visible = TRUE;
				object.enabled = TRUE;
				object.alpha = 255;
				object.alpha_over = 255;
				object.bgcolour = ENG_UNDEFINED;
				object.bgcolour_over = 0x000000;
				object.w = 24;
				object.h = 12;
				strcpy(object.img_file, ENG_IMG_CPS_TEXT);
				object.img_set_colour_key = TRUE;
				object.img_x = 0;
				object.img_y = 0;
				object.img_w = object.w;
				object.img_h = object.h;
				strcpy(object.img_over_file, ENG_IMG_CPS_TEXT);
				object.img_over_x = 0;
				object.img_over_y = 0;
				object.img_over_w = object.w;
				object.img_over_h = object.h;
				object.draggable = ENG_DRAG_RESTRICT_TO_PARENT;
				object.source = ENG_OBJECT_SOURCE_SYSTEM;
				retval = ENG_CreateObject(&object, FALSE);
			}
			if (!retval) {
				/* Create a pointer shadow */
				ENG_SetObjectDefaults(&object);
				object.id = ENG_REC_POINTER;
				object.pid = ENG_LYR_POINTER;
				object.alpha = ENG_REC_POINTER_ALPHA;
				object.alpha_over = ENG_REC_POINTER_ALPHA;
				object.x = ENG_REC_POINTER_X;
				object.y = ENG_REC_POINTER_Y;
				object.w = ENG_IMG_POINTER_W;
				object.h = ENG_IMG_POINTER_H;
				strcpy(object.img_file, ENG_IMG_POINTER_SHADOW);
				object.img_set_colour_key = TRUE;
				strcpy(object.img_over_file, ENG_IMG_POINTER_SHADOW);
				object.navigable = FALSE;
				object.source = ENG_OBJECT_SOURCE_SYSTEM;
				retval = ENG_CreateObject(&object, FALSE);
			}
			if (!retval) {
				/* Create a pointer */
				ENG_SetObjectDefaults(&object);
				object.id = ENG_LYR_POINTER;
				object.pid = ENG_LYR_SCREEN;
				object.klasse = ENG_OBJECT_KLASSE_LAYER;
				object.visible = eng_op_env.pointer;
				object.enabled = TRUE;
				object.alpha = 255;
				object.alpha_over = 255;
				object.w = ENG_IMG_POINTER_W;
				object.h = ENG_IMG_POINTER_H;
				strcpy(object.img_file, ENG_IMG_POINTER);
				object.img_set_colour_key = TRUE;
				strcpy(object.img_over_file, ENG_IMG_POINTER);
				object.navigable = FALSE;
				object.source = ENG_OBJECT_SOURCE_SYSTEM;
				retval = ENG_CreateObject(&object, FALSE);
			}
			if (!retval) {
				/* Create a font object */
				ENG_SetObjectDefaults(&object);
				object.id = ENG_FNT_ASCII_8X12_WHITE;
				object.klasse = ENG_OBJECT_KLASSE_FONT;
				object.visible = FALSE;
				object.w = 760;
				object.h = 13;
				strcpy(object.img_file, ENG_COMMON_IMAGE_DIR ENG_IMG_FONT_ASCII_8X12_WHITE);
				object.img_set_colour_key = TRUE;
				object.source = ENG_OBJECT_SOURCE_SYSTEM;
				retval = ENG_CreateObject(&object, TRUE);
			}
			if (!retval) {
				/* Create a font object */
				ENG_SetObjectDefaults(&object);
				object.id = ENG_FNT_ASCII_8X12_BLACK;
				object.klasse = ENG_OBJECT_KLASSE_FONT;
				object.visible = FALSE;
				object.w = 760;
				object.h = 13;
				strcpy(object.img_file, ENG_COMMON_IMAGE_DIR ENG_IMG_FONT_ASCII_8X12_BLACK);
				object.img_set_colour_key = TRUE;
				object.source = ENG_OBJECT_SOURCE_SYSTEM;
				retval = ENG_CreateObject(&object, TRUE);
			}
			if (!retval) {
				/* Create a font object */
				ENG_SetObjectDefaults(&object);
				object.id = ENG_FNT_ASCII_8X12_WHITE_BLACK_OUTLINE;
				object.klasse = ENG_OBJECT_KLASSE_FONT;
				object.visible = FALSE;
				object.w = 760;
				object.h = 13;
				strcpy(object.img_file, ENG_COMMON_IMAGE_DIR ENG_IMG_FONT_ASCII_8X12_WHITE_BLACK_OUTLINE);
				object.img_set_colour_key = TRUE;
				object.source = ENG_OBJECT_SOURCE_SYSTEM;
				retval = ENG_CreateObject(&object, TRUE);
			}
			if (!retval) {
				/* Create a font object */
				ENG_SetObjectDefaults(&object);
				object.id = ENG_FNT_ASCII_16X24_WHITE;
				object.klasse = ENG_OBJECT_KLASSE_FONT;
				object.visible = FALSE;
				object.w = 1520;
				object.h = 25;
				strcpy(object.img_file, ENG_COMMON_IMAGE_DIR ENG_IMG_FONT_ASCII_16X24_WHITE);
				object.img_set_colour_key = TRUE;
				object.source = ENG_OBJECT_SOURCE_SYSTEM;
				retval = ENG_CreateObject(&object, TRUE);
			}
			if (!retval) {
				/* Create a font object */
				ENG_SetObjectDefaults(&object);
				object.id = ENG_FNT_ASCII_16X24_BLACK;
				object.klasse = ENG_OBJECT_KLASSE_FONT;
				object.visible = FALSE;
				object.w = 1520;
				object.h = 25;
				strcpy(object.img_file, ENG_COMMON_IMAGE_DIR ENG_IMG_FONT_ASCII_16X24_BLACK);
				object.img_set_colour_key = TRUE;
				object.source = ENG_OBJECT_SOURCE_SYSTEM;
				retval = ENG_CreateObject(&object, TRUE);
			}
			if (!retval) {
				/* Create a font object */
				ENG_SetObjectDefaults(&object);
				object.id = ENG_FNT_ASCII_16X24_WHITE_BLACK_OUTLINE;
				object.klasse = ENG_OBJECT_KLASSE_FONT;
				object.visible = FALSE;
				object.w = 1520;
				object.h = 25;
				strcpy(object.img_file, ENG_COMMON_IMAGE_DIR ENG_IMG_FONT_ASCII_16X24_WHITE_BLACK_OUTLINE);
				object.img_set_colour_key = TRUE;
				object.source = ENG_OBJECT_SOURCE_SYSTEM;
				retval = ENG_CreateObject(&object, TRUE);
			}
		}
			
		if (!retval) {
			/* Load the user objects from the objects file */
			printf("%s: Reading from %s\n", __func__, eng_op_env.objects_file);
			/* Open objects file */
			if ((fp = fopen(eng_op_env.objects_file, "r")) == NULL) {
				sprintf(eng_last_error, "%s: %s not found", __func__, eng_op_env.objects_file);
				retval = ENG_ERROR_CANNOT_OPEN_OBJECTS_FILE;
			} else {
				/* Read data until feof and the last object has been created */
				while (!quit) {
					/* Read the whole line and chop off the [CR]LFs */
					line[0] = 0;
					fgets(line, 256, fp);	/* Reads the string + one or more [CR]LFs */
					/* MacOSX and *nix is LF, Win and Mac is CRLF */
					for (count = strlen(line) - 1; count >= 0; count--) {
						if (line[count] == 13 || line[count] == 10) line[count] = 0;
					}
					linecount++;
					#ifdef DEBUG_OBJECTS_FILE
						printf("line: %s\n", line);
					#endif
					if (!strncmp(line, "id=", 3)) {
						strcpy(line, &line[3]);
						if ((sscanf(line, "%i", &object.id)) != 1) object.id = ENG_UNDEFINED;
					} else if (!strncmp(line, "pid=", 4)) {
						strcpy(line, &line[4]);
						if (!strcmp(line, "ENG_LYR_CANVAS")) {
							object.pid = ENG_LYR_CANVAS;
						} else if (!strcmp(line, "ENG_LYR_SCREEN")) {
							object.pid = ENG_LYR_SCREEN;
						} else if (!strcmp(line, "ENG_LYR_VIEWPORT")) {
							object.pid = ENG_LYR_VIEWPORT;
						} else if (!strcmp(line, "ENG_LYR_DEBUG")) {
							object.pid = ENG_LYR_DEBUG;
						} else if (!strcmp(line, "ENG_LYR_POINTER")) {
							object.pid = ENG_LYR_POINTER;
						} else if (!strcmp(line, "ENG_REC_POINTER")) {
							object.pid = ENG_REC_POINTER;
						} else if (strlen(line) > 0) {
							if ((sscanf(line, "%i", &object.pid)) != 1) object.pid = ENG_UNDEFINED;
						}
					} else if (!strncmp(line, "klasse=", 7)) {
						strcpy(line, &line[7]);
						if (!strcmp(line, "ENG_OBJECT_KLASSE_LAYER")) {
							object.klasse = ENG_OBJECT_KLASSE_LAYER;
						} else if (!strcmp(line, "ENG_OBJECT_KLASSE_RECTANGLE")) {
							object.klasse = ENG_OBJECT_KLASSE_RECTANGLE;
						} else if (!strcmp(line, "ENG_OBJECT_KLASSE_FONT")) {
							object.klasse = ENG_OBJECT_KLASSE_FONT;
						} else if (!strcmp(line, "ENG_OBJECT_KLASSE_TEXT")) {
							object.klasse = ENG_OBJECT_KLASSE_TEXT;
						} else if (strlen(line) > 0) {
							if ((sscanf(line, "%i", &object.klasse)) != 1) object.klasse = ENG_UNDEFINED;
						}
					} else if (!strncmp(line, "visible=", 8)) {
						strcpy(line, &line[8]);
						if (!strcmp(line, "ENG_INHERITED")) {
							object.visible = ENG_INHERITED;
						} else if (strlen(line) > 0) {
							if ((sscanf(line, "%i", &object.visible)) != 1) object.visible = ENG_UNDEFINED;
						}
					} else if (!strncmp(line, "enabled=", 8)) {
						strcpy(line, &line[8]);
						if (!strcmp(line, "ENG_INHERITED")) {
							object.enabled = ENG_INHERITED;
						} else if (strlen(line) > 0) {
							if ((sscanf(line, "%i", &object.enabled)) != 1) object.enabled = ENG_UNDEFINED;
						}
					} else if (!strncmp(line, "alpha=", 6)) {
						strcpy(line, &line[6]);
						if (!strcmp(line, "ENG_INHERITED")) {
							object.alpha = ENG_INHERITED;
						} else if (strlen(line) > 0) {
							if ((sscanf(line, "%i", &object.alpha)) != 1) object.alpha = ENG_UNDEFINED;
						}
					} else if (!strncmp(line, "alpha_over=", 11)) {
						strcpy(line, &line[11]);
						if (!strcmp(line, "ENG_INHERITED")) {
							object.alpha_over = ENG_INHERITED;
						} else if (strlen(line) > 0) {
							if ((sscanf(line, "%i", &object.alpha_over)) != 1) object.alpha_over = ENG_UNDEFINED;
						}
					} else if (!strncmp(line, "bgcolour=", 9)) {
						strcpy(line, &line[9]);
						if (!strcmp(line, "ENG_UNDEFINED")) {
							object.bgcolour = ENG_UNDEFINED;
						} else if (strlen(line) > 0) {
							if ((sscanf(line, "%i", &object.bgcolour)) != 1) object.bgcolour = ENG_UNDEFINED - 1;
						}
					} else if (!strncmp(line, "bgcolour_over=", 14)) {
						strcpy(line, &line[14]);
						if (!strcmp(line, "ENG_UNDEFINED")) {
							object.bgcolour_over = ENG_UNDEFINED;
						} else if (strlen(line) > 0) {
							if ((sscanf(line, "%i", &object.bgcolour_over)) != 1) object.bgcolour_over = ENG_UNDEFINED - 1;
						}
					} else if (!strncmp(line, "x=", 2)) {
						strcpy(line, &line[2]);
						if (strlen(line) > 0) {
							if ((sscanf(line, "%i", &object.x)) != 1) object.x = 0;
						}
					} else if (!strncmp(line, "y=", 2)) {
						strcpy(line, &line[2]);
						if (strlen(line) > 0) {
							if ((sscanf(line, "%i", &object.y)) != 1) object.y = 0;
						}
					} else if (!strncmp(line, "w=", 2)) {
						strcpy(line, &line[2]);
						if (!strcmp(line, "screen_w")) {
							object.w = eng_op_env.screen_w;
						} else if (strlen(line) > 0) {
							if ((sscanf(line, "%i", &object.w)) != 1) object.w = 0;
						}
					} else if (!strncmp(line, "h=", 2)) {
						strcpy(line, &line[2]);
						if (!strcmp(line, "screen_h")) {
							object.h = eng_op_env.screen_h;
						} else if (strlen(line) > 0) {
							if ((sscanf(line, "%i", &object.h)) != 1) object.h = 0;
						}
					} else if (!strncmp(line, "img_file=", 9)) {
						strcpy(line, &line[9]);
						if (strlen(line) >= ENG_FILE_LENGTH) {
							sprintf(eng_last_error, "%s: img_file property in %s for object at line %i is too long (max %i)", __func__, eng_op_env.objects_file, objectline, ENG_FILE_LENGTH - 1);
							retval = ENG_ERROR_OBJECTS_FILE_SYNTAX_ERROR;
							break;
						} else if (strlen(line) > 0) {
							strcpy(object.img_file, line);
						}
					} else if (!strncmp(line, "img_set_colour_key=", 19)) {
						strcpy(line, &line[19]);
						if (strlen(line) > 0) {
							if ((sscanf(line, "%i", &object.img_set_colour_key)) != 1) object.img_set_colour_key = ENG_UNDEFINED;
						}
					} else if (!strncmp(line, "img_use_rleaccel=", 17)) {
						strcpy(line, &line[17]);
						if (strlen(line) > 0) {
							if ((sscanf(line, "%i", &object.img_use_rleaccel)) != 1) object.img_use_rleaccel = ENG_UNDEFINED;
						}
					} else if (!strncmp(line, "img_x=", 6)) {
						strcpy(line, &line[6]);
						if (strlen(line) > 0) {
							if ((sscanf(line, "%i", &object.img_x)) != 1) object.img_x = ENG_UNDEFINED;
						}
					} else if (!strncmp(line, "img_y=", 6)) {
						strcpy(line, &line[6]);
						if (strlen(line) > 0) {
							if ((sscanf(line, "%i", &object.img_y)) != 1) object.img_y = ENG_UNDEFINED;
						}
					} else if (!strncmp(line, "img_w=", 6)) {
						strcpy(line, &line[6]);
						if (!strcmp(line, "screen_w")) {
							object.img_w = eng_op_env.screen_w;
						} else if (strlen(line) > 0) {
							if ((sscanf(line, "%i", &object.img_w)) != 1) object.img_w = ENG_UNDEFINED;
						}
					} else if (!strncmp(line, "img_h=", 6)) {
						strcpy(line, &line[6]);
						if (!strcmp(line, "screen_h")) {
							object.img_h = eng_op_env.screen_h;
						} else if (strlen(line) > 0) {
							if ((sscanf(line, "%i", &object.img_h)) != 1) object.img_h = ENG_UNDEFINED;
						}
					} else if (!strncmp(line, "img_over_file=", 14)) {
						strcpy(line, &line[14]);
						if (strlen(line) >= ENG_FILE_LENGTH) {
							sprintf(eng_last_error, "%s: img_over_file property in %s for object at line %i is too long (max %i)", __func__, eng_op_env.objects_file, objectline, ENG_FILE_LENGTH - 1);
							retval = ENG_ERROR_OBJECTS_FILE_SYNTAX_ERROR;
							break;
						} else if (strlen(line) > 0) {
							strcpy(object.img_over_file, line);
						}
					} else if (!strncmp(line, "img_over_set_colour_key=", 24)) {
						strcpy(line, &line[24]);
						if (strlen(line) > 0) {
							if ((sscanf(line, "%i", &object.img_over_set_colour_key)) != 1) object.img_over_set_colour_key = ENG_UNDEFINED;
						}
					} else if (!strncmp(line, "img_over_use_rleaccel=", 22)) {
						strcpy(line, &line[22]);
						if (strlen(line) > 0) {
							if ((sscanf(line, "%i", &object.img_over_use_rleaccel)) != 1) object.img_over_use_rleaccel = ENG_UNDEFINED;
						}
					} else if (!strncmp(line, "img_over_x=", 11)) {
						strcpy(line, &line[11]);
						if (strlen(line) > 0) {
							if ((sscanf(line, "%i", &object.img_over_x)) != 1) object.img_over_x = ENG_UNDEFINED;
						}
					} else if (!strncmp(line, "img_over_y=", 11)) {
						strcpy(line, &line[11]);
						if (strlen(line) > 0) {
							if ((sscanf(line, "%i", &object.img_over_y)) != 1) object.img_over_y = ENG_UNDEFINED;
						}
					} else if (!strncmp(line, "img_over_w=", 11)) {
						strcpy(line, &line[11]);
						if (!strcmp(line, "screen_w")) {
							object.img_over_w = eng_op_env.screen_w;
						} else if (strlen(line) > 0) {
							if ((sscanf(line, "%i", &object.img_over_w)) != 1) object.img_over_w = ENG_UNDEFINED;
						}
					} else if (!strncmp(line, "img_over_h=", 11)) {
						strcpy(line, &line[11]);
						if (!strcmp(line, "screen_h")) {
							object.img_over_h = eng_op_env.screen_h;
						} else if (strlen(line) > 0) {
							if ((sscanf(line, "%i", &object.img_over_h)) != 1) object.img_over_h = ENG_UNDEFINED;
						}
					} else if (!strncmp(line, "draggable=", 10)) {
						strcpy(line, &line[10]);
						if (!strcmp(line, "ENG_DRAG_ANYWHERE")) {
							object.draggable = ENG_DRAG_ANYWHERE;
						} else if (!strcmp(line, "ENG_DRAG_RESTRICT_TO_PARENT")) {
							object.draggable = ENG_DRAG_RESTRICT_TO_PARENT;
						} else if (strlen(line) > 0) {
							if ((sscanf(line, "%i", &object.draggable)) != 1) object.draggable = ENG_UNDEFINED;
						}
					} else if (!strncmp(line, "drag_point=", 11)) {
						strcpy(line, &line[11]);
						if (!strcmp(line, "ENG_DRAG_POINT_ANYWHERE")) {
							object.drag_point = ENG_DRAG_POINT_ANYWHERE;
						} else if (!strcmp(line, "ENG_DRAG_POINT_MIDDLE")) {
							object.drag_point = ENG_DRAG_POINT_MIDDLE;
						} else if (strlen(line) > 0) {
							if ((sscanf(line, "%i", &object.drag_point)) != 1) object.drag_point = ENG_UNDEFINED;
						}
					} else if (!strncmp(line, "navigable=", 10)) {
						strcpy(line, &line[10]);
						if (strlen(line) > 0) {
							if ((sscanf(line, "%i", &object.navigable)) != 1) object.navigable = ENG_UNDEFINED;
						}
					} else if ((!strncmp(line, "[object]", 8)) || (feof(fp))) {
						if (!setdefaults) {
							/* Create object */
							object.source = ENG_OBJECT_SOURCE_FILE;
							strcpy(object.objects_file, eng_op_env.objects_file);
							retval = ENG_CreateObject(&object, FALSE);
							if (retval) {
								sprintf(last_error, " in %s for object at line %i", eng_op_env.objects_file, objectline);
								strcat(eng_last_error, last_error);
								break;
							}
							loadcount++;
							setdefaults = TRUE;
						}
						if (setdefaults) {
							/* Set some default object properties */
							ENG_SetObjectDefaults(&object);
							setdefaults = FALSE;
						}
						objectline = linecount;
						if (feof(fp)) quit = TRUE;
					}
				}
				fclose(fp);
			}
		}

		if (!retval) {
			/* [Re]set some important variables */
			eng_oup = eng_lyr_root;

			/* Display the loaded object count */
			printf("%s: Loaded %i objects\n", __func__, loadcount);

			#ifdef DEBUG_OBJECT_SORT
				printf("%s: Sorted eng_objects_id :-\n", __func__);
				for (count = 0; count < eng_object_count; count++) {
					printf("    %p: id=%i pob=%p zorder=%i\n", eng_objects_id[count], eng_objects_id[count]->id, eng_objects_id[count]->pob, eng_objects_id[count]->zorder);
				}
			#endif
			#ifdef DEBUG_IMAGE_SORT
				printf("%s: Sorted eng_images :-\n", __func__);
				for (count = 0; count < eng_image_count; count++) {
					printf("    %p: img_file=%s img_surface=%p\n", eng_images[count], eng_images[count]->img_file, eng_images[count]->img_surface;
				}
			#endif

			if (send_post_event) {
				/* The user may want to do something on a successful
				   load before entering/returning to the main loop */
				event.object = eng_lyr_root; event.klasse = ENG_EVENT_POST_LOAD_OBJECTS; ENG_SendEvent(&event);
			}
		}
	}

	return retval;
}

/***************************************************************************
 * Get Object                                                              *
 ***************************************************************************/
/* This uses the system maintained eng_objects_id array to binary search
   for the object id */

/* On entry: id is the object id
	On exit: returns a pointer to the object on success else NULL */

ENG_Object *ENG_GetObject(int id) {
	int top = eng_object_count - 1, middle = 0, bottom = 0;

	if (eng_started == FALSE) {
		sprintf(eng_last_error, "%s: ENG has not been started yet", __func__);
	} else {
		/* Binary search for id */
		while (bottom <= top) {
			middle = (bottom + top) / 2;
			if (eng_objects_id[middle]->id == id) {
				return eng_objects_id[middle];
			} else if (eng_objects_id[middle]->id < id) {
				bottom = middle + 1;
			} else {
				top = middle - 1;
			}
		}

		/* int count;

		/$ Linear search for id $/
		for (count = 0; count < eng_object_count; count++) {
			if (eng_objects[count]->id == id) return eng_objects[count];
		} */
	}

	return NULL;
}

/***************************************************************************
 * Create Object                                                           *
 ***************************************************************************/
/* This will create the object and create any declared images.
   There may be existing objects within eng_objects whose pob is NULL because
   their parent was recently freed or hasn't been created yet. This function
   will attempt to fix these NULL pobs on introduction of this new object.
   Additionally their may be objects whose img_surface and/or img_over_surface
   are NULL and these will be fixed too */
   
/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_CreateObject(ENG_Object *object, int ignore_image_dir) {
	int retval = ENG_ERROR_NONE, swapped, count;
	ENG_Object *newobject, *swap_object;
	ENG_Image image;
	
	if (eng_started == FALSE) {
		sprintf(eng_last_error, "%s: ENG has not been started yet", __func__);
		retval = ENG_ERROR_ENG_NOT_YET_STARTED;
	} else  if (eng_object_count >= ENG_MAX_OBJECTS) {
		sprintf(eng_last_error, "%s: Too many objects in %s (max %i)", __func__, eng_op_env.objects_file, ENG_MAX_OBJECTS);
		retval = ENG_ERROR_TOO_MANY_OBJECTS;
	} else {
		/* Validate data */
		/* Check uniqueness of id */
		for (count = 0; count < eng_object_count; count++) {
			if (object->id == eng_objects[count]->id) {
				sprintf(eng_last_error, "%s: Invalid duplicate id property", __func__);
				retval = ENG_ERROR_OBJECTS_FILE_SYNTAX_ERROR;
				break;
			}
		}
		if (!retval) {
			/* Check the rest */
			if (object->id < 0) {
				sprintf(eng_last_error, "%s: Invalid id property", __func__);
				retval = ENG_ERROR_OBJECTS_FILE_SYNTAX_ERROR;
			} else if (object->pid < 0) {
				sprintf(eng_last_error, "%s: Invalid pid property", __func__);
				retval = ENG_ERROR_OBJECTS_FILE_SYNTAX_ERROR;
			} else if (object->klasse < ENG_OBJECT_KLASSE_RECTANGLE || object->klasse > ENG_OBJECT_KLASSE_TEXT) {
				sprintf(eng_last_error, "%s: Invalid klasse property", __func__);
				retval = ENG_ERROR_OBJECTS_FILE_SYNTAX_ERROR;
			} else if (object->visible != 0 && object->visible != 1 && object->visible != ENG_INHERITED) {
				sprintf(eng_last_error, "%s: Invalid visible property", __func__);
				retval = ENG_ERROR_OBJECTS_FILE_SYNTAX_ERROR;
			} else if (object->enabled != 0 && object->enabled != 1 && object->enabled != ENG_INHERITED) {
				sprintf(eng_last_error, "%s: Invalid enabled property", __func__);
				retval = ENG_ERROR_OBJECTS_FILE_SYNTAX_ERROR;
			} else if ((object->alpha < 0 || object->alpha > 255) && object->alpha != ENG_INHERITED) {
				sprintf(eng_last_error, "%s: Invalid alpha property", __func__);
				retval = ENG_ERROR_OBJECTS_FILE_SYNTAX_ERROR;
			} else if ((object->alpha_over < 0 || object->alpha_over > 255) && object->alpha_over != ENG_INHERITED) {
				sprintf(eng_last_error, "%s: Invalid alpha_over property", __func__);
				retval = ENG_ERROR_OBJECTS_FILE_SYNTAX_ERROR;
			} else if (object->bgcolour < ENG_UNDEFINED || object->bgcolour > 0xFFFFFF) {
				sprintf(eng_last_error, "%s: Invalid bgcolour property", __func__);
				retval = ENG_ERROR_OBJECTS_FILE_SYNTAX_ERROR;
			} else if (object->bgcolour_over < ENG_UNDEFINED || object->bgcolour_over > 0xFFFFFF) {
				sprintf(eng_last_error, "%s: Invalid bgcolour_over property", __func__);
				retval = ENG_ERROR_OBJECTS_FILE_SYNTAX_ERROR;
			} else if (object->img_set_colour_key < 0 || object->img_set_colour_key > 1) {
				sprintf(eng_last_error, "%s: Invalid img_set_colour_key property", __func__);
				retval = ENG_ERROR_OBJECTS_FILE_SYNTAX_ERROR;
			} else if (object->img_use_rleaccel < 0 || object->img_use_rleaccel > 1) {
				sprintf(eng_last_error, "%s: Invalid img_use_rleaccel property", __func__);
				retval = ENG_ERROR_OBJECTS_FILE_SYNTAX_ERROR;
			} else if (object->img_over_set_colour_key < 0 || object->img_over_set_colour_key > 1) {
				sprintf(eng_last_error, "%s: Invalid img_over_set_colour_key property", __func__);
				retval = ENG_ERROR_OBJECTS_FILE_SYNTAX_ERROR;
			} else if (object->img_over_use_rleaccel < 0 || object->img_over_use_rleaccel > 1) {
				sprintf(eng_last_error, "%s: Invalid img_over_use_rleaccel property", __func__);
				retval = ENG_ERROR_OBJECTS_FILE_SYNTAX_ERROR;
			} else if (object->draggable < 0 || object->draggable > ENG_DRAG_RESTRICT_TO_PARENT) {
				sprintf(eng_last_error, "%s: Invalid draggable property", __func__);
				retval = ENG_ERROR_OBJECTS_FILE_SYNTAX_ERROR;
			} else if (object->drag_point < ENG_DRAG_POINT_ANYWHERE || object->drag_point > ENG_DRAG_POINT_MIDDLE) {
				sprintf(eng_last_error, "%s: Invalid drag_point property", __func__);
				retval = ENG_ERROR_OBJECTS_FILE_SYNTAX_ERROR;
			} else if (object->navigable < 0 || object->navigable > 1) {
				sprintf(eng_last_error, "%s: Invalid navigable property", __func__);
				retval = ENG_ERROR_OBJECTS_FILE_SYNTAX_ERROR;
			}
		}
		if (!retval) {
			/* Allocate memory for the new object */
			newobject = (ENG_Object*) malloc(sizeof(ENG_Object));
			*newobject = *object;
			#ifdef DEBUG_OBJECTS_ALL_DRAGGABLE
				if (newobject->draggable == FALSE) newobject->draggable = ENG_DRAG_ANYWHERE;
			#endif
			#ifdef DEBUG_OBJECTS
				printf("%s: Creating %p (%i)\n", __func__, newobject, newobject->id);
			#endif
			/* Store the object pointer in the arrays */
			eng_objects[eng_object_count] = newobject;
			eng_objects_id[eng_object_count] = newobject;
			/* Set the zorder */
			newobject->zorder = eng_object_count;
			/* Update the array element count now */
			eng_object_count++;
			/* Store pointers to system objects */
			if (newobject->id == ENG_LYR_ROOT) {
				eng_lyr_root = newobject;
			} else if (newobject->id == ENG_LYR_CANVAS) {
				eng_lyr_canvas = newobject;
			} else if (newobject->id == ENG_LYR_SCREEN) {
				eng_lyr_screen = newobject;
			} else if (newobject->id == ENG_LYR_VIEWPORT) {
				eng_lyr_viewport = newobject;
			} else if (newobject->id == ENG_LYR_DEBUG) {
				eng_lyr_debug = newobject;
			} else if (newobject->id == ENG_REC_POINTER) {
				eng_rec_pointer = newobject;
			} else if (newobject->id == ENG_LYR_POINTER) {
				eng_lyr_pointer = newobject;
			}
			/* Create img_file if declared */
			if (strlen(newobject->img_file) > 0) {
				strcpy(image.img_file, newobject->img_file);
				image.img_surface = NULL;
				/* If image hasn't already been created then create it */
				if (ENG_GetImage(image.img_file) == NULL) {
					retval = ENG_CreateImage(&image, ignore_image_dir, newobject->w, newobject->h, newobject->img_set_colour_key, newobject->img_use_rleaccel);
					if (!retval) {
						/* Set img_surface */
						newobject->img_surface = ENG_GetImage(newobject->img_file)->img_surface;
					} else {
						/* No img_file so free object (ignoring errors) */
						ENG_FreeObject(newobject, FALSE);
					}
				}
			}
			if (!retval) {
				/* Create img_over_file if declared */
				if (strlen(newobject->img_over_file) > 0) {
					strcpy(image.img_file, newobject->img_over_file);
					image.img_surface = NULL;
					/* If image hasn't already been created then create it */
					if (ENG_GetImage(image.img_file) == NULL) {
						retval = ENG_CreateImage(&image, ignore_image_dir, newobject->w, newobject->h, newobject->img_over_set_colour_key, newobject->img_over_use_rleaccel);
						if (!retval) {
							/* Set img_over_surface */
							newobject->img_over_surface = ENG_GetImage(newobject->img_over_file)->img_surface;
						} else {
							/* No img_over_file so free object (ignoring errors) */
							ENG_FreeObject(newobject, FALSE);
						}
					}
				}
			}
			if (!retval) {
				/* Bubble sort eng_objects_id for later binary search */
				do {
					swapped = FALSE;
					for (count = 0; count < eng_object_count - 1; count++) {
						if (eng_objects_id[count]->id > eng_objects_id[count + 1]->id) {
							swapped = TRUE;
							swap_object = eng_objects_id[count];
							eng_objects_id[count] = eng_objects_id[count + 1];
							eng_objects_id[count + 1] = swap_object;
						}
					}
				} while (swapped);
				/* Iterate through eng_objects and fix the pob, img_surface
				   and img_over_surface properties that are NULL */
				for (count = 0; count < eng_object_count; count++) {
					if (eng_objects[count]->pob == NULL) {
						/* This may still result in a NULL pipd if the parent never exists */
						eng_objects[count]->pob = ENG_GetObject(eng_objects[count]->pid);
					}
					if (strlen(eng_objects[count]->img_file) > 0 && eng_objects[count]->img_surface == NULL) {
						/* This may still result in a NULL img_surface if the image never exists */
						eng_objects[count]->img_surface = ENG_GetImage(eng_objects[count]->img_file)->img_surface;
					}
					if (strlen(eng_objects[count]->img_over_file) > 0 && eng_objects[count]->img_over_surface == NULL) {
						/* This may still result in a NULL img_over_surface if the image never exists */
						eng_objects[count]->img_over_surface = ENG_GetImage(eng_objects[count]->img_over_file)->img_surface;
					}
				}
				/* Maintain the highest zorder for any existing pointer, shadow and debug window */
				if (eng_lyr_pointer != NULL && eng_rec_pointer != NULL) {
					ENG_SetObjectZOrder(eng_lyr_pointer, ENG_ZORDER_ABOVE, newobject);
					ENG_SetObjectZOrder(eng_rec_pointer, ENG_ZORDER_BELOW, eng_lyr_pointer);
					ENG_SetObjectZOrder(eng_lyr_debug, ENG_ZORDER_BELOW, eng_rec_pointer);
				}
				#ifdef DEBUG_OBJECTS
					printf("newobject :-\n");
					printf("id=%i\n", newobject->id);
					printf("pid=%i\n", newobject->pid);
					printf("klasse=%i\n", newobject->klasse);
					printf("visible=%i\n", newobject->visible);
					printf("enabled=%i\n", newobject->enabled);
					printf("alpha=%i\n", newobject->alpha);
					printf("alpha_over=%i\n", newobject->alpha_over);
					printf("bgcolour=%03X\n", newobject->bgcolour);
					printf("bgcolour_over=%03X\n", newobject->bgcolour_over);
					printf("x=%i\n", newobject->x);
					printf("y=%i\n", newobject->y);
					printf("w=%i\n", newobject->w);
					printf("h=%i\n", newobject->h);
					printf("img_file=%s\n", newobject->img_file);
					printf("img_set_colour_key=%i\n", newobject->img_set_colour_key);
					printf("img_use_rleaccel=%i\n", newobject->img_use_rleaccel);
					printf("img_x=%i\n", newobject->img_x);
					printf("img_y=%i\n", newobject->img_y);
					printf("img_w=%i\n", newobject->img_w);
					printf("img_h=%i\n", newobject->img_h);
					printf("img_over_file=%s\n", newobject->img_over_file);
					printf("img_over_set_colour_key=%i\n", newobject->img_over_set_colour_key);
					printf("img_over_use_rleaccel=%i\n", newobject->img_over_use_rleaccel);
					printf("img_over_x=%i\n", newobject->img_over_x);
					printf("img_over_y=%i\n", newobject->img_over_y);
					printf("img_over_w=%i\n", newobject->img_over_w);
					printf("img_over_h=%i\n", newobject->img_over_h);
					printf("draggable=%i\n", newobject->draggable);
					printf("drag_point=%i\n", newobject->drag_point);
					printf("navigable=%i\n", newobject->navigable);
					printf("zorder=%i\n", newobject->zorder);
					printf("pob=%p\n", newobject->pob);
					printf("img_surface=%p\n", newobject->img_surface);
					printf("img_over_surface=%p\n", newobject->img_over_surface);
					printf("source=%i\n", newobject->source);
					printf("objects_file=%s\n", newobject->objects_file);
				#endif
			}
		}
	}

	return retval;
}

/***************************************************************************
 * Set Object Defaults                                                     *
 ***************************************************************************/
/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_SetObjectDefaults(ENG_Object *object) {
	int retval = ENG_ERROR_NONE;

	if (object == NULL || sizeof(*object) != sizeof(ENG_Object)) {
		sprintf(eng_last_error, "%s: object is invalid", __func__);
		retval = ENG_ERROR_INVALID_OBJECT;
	} else {
		object->id = ENG_UNDEFINED;
		object->pid = ENG_LYR_CANVAS;
		object->klasse = ENG_OBJECT_KLASSE_RECTANGLE;
		object->visible = ENG_INHERITED;
		object->enabled = ENG_INHERITED;
		object->alpha = ENG_INHERITED;
		object->alpha_over = ENG_INHERITED;
		object->bgcolour = ENG_UNDEFINED;
		object->bgcolour_over = ENG_UNDEFINED;
		object->x = 0;
		object->y = 0;
		object->w = 0;
		object->h = 0;
		strcpy(object->img_file, "");
		object->img_set_colour_key = 0;
		object->img_use_rleaccel = 0;
		object->img_x = ENG_UNDEFINED;
		object->img_y = ENG_UNDEFINED;
		object->img_w = ENG_UNDEFINED;
		object->img_h = ENG_UNDEFINED;
		strcpy(object->img_over_file, "");
		object->img_over_set_colour_key = 0;
		object->img_over_use_rleaccel = 0;
		object->img_over_x = ENG_UNDEFINED;
		object->img_over_y = ENG_UNDEFINED;
		object->img_over_w = ENG_UNDEFINED;
		object->img_over_h = ENG_UNDEFINED;
		object->draggable = FALSE;
		object->drag_point = ENG_DRAG_POINT_ANYWHERE;
		object->navigable = TRUE;
		object->zorder = ENG_UNDEFINED;
		object->pob = NULL;
		object->img_surface = NULL;
		object->img_over_surface = NULL;
		object->source = ENG_OBJECT_SOURCE_USER;
		strcpy(object->objects_file, "");
	}
	
	return retval;
}

/***************************************************************************
 * Free Object                                                             *
 ***************************************************************************/
/* When freeing a parent object, the child objects will have their pob
   nullified up until the point when (or if) the parent is reinstated.
   Images declared within the object are freed and any objects that are
   also referencing these images through img_surface and img_over_surface
   will have these properties nullified */

/* On entry: object is the object to free
			 free_images_only = TRUE to free ONLY any declared images */
/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_FreeObject(ENG_Object *object, int free_images_only) {
	int retval = ENG_ERROR_NONE, count, count2;
	ENG_Image *image;
	SDL_Surface *img_surface;

	if (eng_object_count <= 0) {
		sprintf(eng_last_error, "%s: There is no object to free", __func__);
		retval = ENG_ERROR_INVALID_OBJECT;
	} else if (object == NULL || sizeof(*object) != sizeof(ENG_Object)) {
		sprintf(eng_last_error, "%s: object is invalid", __func__);
		retval = ENG_ERROR_INVALID_OBJECT;
	} else if (eng_run && object->source == ENG_OBJECT_SOURCE_SYSTEM) {
		sprintf(eng_last_error, "%s: object is invalid", __func__);
		retval = ENG_ERROR_INVALID_OBJECT;
	} else {
		#ifdef DEBUG_OBJECTS
			printf("%s: Freeing %p (%i)\n", __func__, object, object->id);
		#endif
		/* Free img_file and img_over_file if either are declared */
		for (count = 0; count < 2; count++) {
			if (count == 0) {
				image = ENG_GetImage(object->img_file);
			} else {
				image = ENG_GetImage(object->img_over_file);
			}
			if (image != NULL) {
				img_surface = image->img_surface;
				if (!ENG_FreeImage(image)) {
					/* Nullify all surface pointers within eng_objects that reference this surface */
					for (count2 = 0; count2 < eng_object_count; count2++) {
						if (eng_objects[count2]->img_surface == img_surface) {
							eng_objects[count2]->img_surface = NULL;
						}
						if (eng_objects[count2]->img_over_surface == img_surface) {
							eng_objects[count2]->img_over_surface = NULL;
						}
					}
				}
			}
		}
		if (!free_images_only) {
			/* If the current object is the parent to other objects then nullify their pob */
			for (count = 0; count < eng_object_count; count++) {
				if (eng_objects[count]->pob == object) eng_objects[count]->pob = NULL;
			}
			/* If anything is above the current object then move it down the array */
			for (count = object->zorder; count < eng_object_count - 1; count++) {
				eng_objects[count] = eng_objects[count + 1];
				eng_objects[count]->zorder--;
			}
			/* Do the same with eng_objects_id */
			for (count = 0; count < eng_object_count - 1; count++) {
				if (eng_objects_id[count] == object) {
					for (count2 = count; count2 < eng_object_count - 1; count2++) {
						eng_objects_id[count2] = eng_objects_id[count2 + 1];
					}
					break;				
				}
			}
			/* Nullify eng_oup if it's the same object */
			if (eng_oup == object) eng_oup = NULL;
			/* Free the memory and update the array element count now */
			free(object);
			eng_object_count--;
		}
	}

	return retval;
}

/***************************************************************************
 * Unload Objects                                                          *
 ***************************************************************************/
/* This enables the user to unload the objects previously loaded from an
   objects file */

int ENG_UnloadObjects(char *objects_file) {
	int retval = ENG_ERROR_NONE, count;
	
	if (eng_started == FALSE) {
		sprintf(eng_last_error, "%s: ENG has not been started yet", __func__);
		retval = ENG_ERROR_ENG_NOT_YET_STARTED;
	} else {
		printf("%s: Unloading %s\n", __func__, objects_file);
		/* Free all objects loaded from objects_file (ignoring errors) */
		for (count = eng_object_count - 1; count >= 0; count--) {
			if (eng_objects[count]->source == ENG_OBJECT_SOURCE_FILE && !strcmp(eng_objects[count]->objects_file, objects_file)) {
				ENG_FreeObject(eng_objects[count], FALSE);
			}
		}
		#ifdef DEBUG_OBJECT_SORT
			printf("%s: Sorted eng_objects_id :-\n", __func__);
			for (count = 0; count < eng_object_count; count++) {
				printf("    %p: id=%i pob=%p zorder=%i\n", eng_objects_id[count], eng_objects_id[count]->id, eng_objects_id[count]->pob, eng_objects_id[count]->zorder);
			}
		#endif
		#ifdef DEBUG_IMAGE_SORT
			printf("%s: Sorted eng_images :-\n", __func__);
			for (count = 0; count < eng_image_count; count++) {
				printf("    %p: img_file=%s img_surface=%p\n", eng_images[count], eng_images[count]->img_file, eng_images[count]->img_surface);
			}
		#endif
	}
	
	return retval;
}

/***************************************************************************
 * Free All Objects                                                        *
 ***************************************************************************/
/* This mops-up any remaining objects before shutting down */

void ENG_FreeAllObjects(void) {
	int count;
	
	if (eng_object_count > 0) {
		printf("%s: Freeing\n", __func__);
		/* Free all remaining objects (ignoring errors) */
		for (count = eng_object_count - 1; count >= 0; count--) {
			ENG_FreeObject(eng_objects[count], FALSE);
		}
	}
}





