/* Demo 01 (c) Thunor 2008 <thunorsif@hotmail.com>
www.thunor.org.uk

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. */


/* Includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "main.h"


/* Defines */
/* It would be more efficient to calculate max lines and malloc the array
   but it's only 500kB and this is a demo. 128000 lines will accommodate
   a 940kB ebook using the 16x24 font and the minimum window width. I used
   The Shape of Things to Come (1933) by H G Wells for testing */
#define EBOOK_MAX_LINES 128000

/* Demo 01 key object ids */
#define LYR_DEMO_01_MAIN 30000
#define LYR_DEMO_01_MINIMISED 30020
#define REC_DEMO_01_WINDOW 30100
#define LYR_DEMO_01_TITLEBAR 30120
#define REC_DEMO_01_TITLEBAR_EXIT 30130
#define REC_DEMO_01_TITLEBAR_MAXIMISE 30132
#define REC_DEMO_01_TITLEBAR_MINIMISE 30134
#define REC_DEMO_01_TITLEBAR_ICON 30136
#define REC_DEMO_01_TOOLBAR 30140
#define REC_DEMO_01_TOOLBAR_FONT_SIZE 30160
#define REC_DEMO_01_TOOLBAR_FONT 30180
#define REC_DEMO_01_TOOLBAR_OPACITY 30220
#define REC_DEMO_01_TOOLBAR_FULLSCREEN 30240
#define REC_DEMO_01_WINDOW_LEFT 30260
#define REC_DEMO_01_WINDOW_RIGHT 30280
#define REC_DEMO_01_STATUSBAR 30300
#define TXT_DEMO_01_STATUSBAR_LINE_STATS 30310
#define TXT_DEMO_01_TEXTAREA 30320
#define REC_DEMO_01_SCROLLBAR_POB 30340
#define REC_DEMO_01_SCROLLBAR_BUTTON_UP 30360
#define REC_DEMO_01_SCROLLBAR_BUTTON_DOWN 30380
#define REC_DEMO_01_SCROLLBAR_UP 30400
#define REC_DEMO_01_SCROLLBAR_DOWN 30420
#define REC_DEMO_01_SCROLLBAR_HANDLE 30440
#define REC_DEMO_01_SHADOW_E 30460
#define REC_DEMO_01_SHADOW_S 30480


/* Data */
char demo_01_ebook_error_text[256];
char *demo_01_ebook_text = NULL;	/* Memory will be allocated to accommodate a file */
char demo_01_line_stats_text[] = "Line xxxxxxxxxx of xxxxxxxxxx";


/* Function prototypes */
int demo_01_load_ebook(void);
void demo_01_expand_everything(void);
void demo_01_init_line_array(int *line_count, int *line_index, char *line_array[], char *current_char, ENG_ASCIIText *ASCIIText_ebook);
void demo_01_calculate_page_lines(ENG_ASCIIText *ASCIIText_ebook, int *page_lines);
void demo_01_refresh_page_navigation_controls(int *line_count, int *line_index, ENG_ASCIIText *ASCIIText_line_stats, int adjust_handle);


/***************************************************************************
 * Demo 01 Event Processor                                                 *
 ***************************************************************************/
/* Expandable ASCII Text Reader */

void demo_01_event_processor(ENG_Event *event) {
	static ENG_Object *demo_main, *minimised, *window, *titlebar, *titlebar_maximise, *textarea, *scrollbar_pob, *scrollbar_up, *scrollbar_down, *scrollbar_handle, *statusbar, *statusbar_line_stats;
	static SDL_Surface *font_ascii_8x12_white, *font_ascii_8x12_black, *font_ascii_8x12_white_black_outline;
	static SDL_Surface *font_ascii_16x24_white, *font_ascii_16x24_black, *font_ascii_16x24_white_black_outline;
	static SDL_Surface *image_maximise, *image_restore;
	static int fs_last_x, fs_last_y, fs_last_w, fs_last_h, fs_last_draggable;
	static int max_last_x, max_last_y, max_last_w, max_last_h, max_last_draggable;
	static int window_fullscreen, window_maximised;
	static ENG_ASCIIText ASCIIText_ebook, ASCIIText_line_stats;
	static char objects_file[ENG_PATH_LENGTH], *line_array[EBOOK_MAX_LINES], *current_char, ebook_text_offset[ENG_CONFIG_OPTION_VALUE_LENGTH];
	static int line_count, line_index, page_lines, ebook_loaded;
	ENG_Object new_object;
	int offset;
		
	/* Process events */
	switch (event->klasse) {
		case ENG_EVENT_OVER:
			break;
		case ENG_EVENT_OUT:
			break;
		case ENG_EVENT_PRESS:
			break;
		case ENG_EVENT_CLICK:
			if (event->object->id == REC_DEMO_01_TITLEBAR_EXIT) {
				/* Save demo_01.ebook_text_offset */
				sprintf(ebook_text_offset, "%i", current_char - demo_01_ebook_text);
				if (ENG_SetConfigOption("demo_01.ebook_text_offset", ebook_text_offset)) printf("%s: %s\n", __func__, ENG_GetError());
				/* Free the memory allocated for the ebook */
				if (ebook_loaded == TRUE) free(demo_01_ebook_text);
				/* Change the event processor to tetravex_3x3 */
				op_env.event_processor = &tetravex_3x3_event_processor;
				if (ENG_SetOpEnvEventProcessor(op_env.event_processor)) printf("%s: %s\n", __func__, ENG_GetError());
				/* Unload the demo_01 objects file */
				if (ENG_UnloadObjects(objects_file)) printf("%s: %s\n", __func__, ENG_GetError());
				/* Free LYR_DEMO_01_MAIN as we created it here */
				if (ENG_FreeObject(demo_main, FALSE)) printf("%s: %s\n", __func__, ENG_GetError());
			} else if (event->object->id == REC_DEMO_01_TITLEBAR_MAXIMISE) {
				/* Toggle window maximised */
				if (window_maximised) {
					window_maximised = FALSE;
					titlebar->x = max_last_x; titlebar->y = max_last_y;
					window->w = max_last_w; window->h = max_last_h;
					titlebar->draggable = max_last_draggable;
					titlebar_maximise->img_surface = titlebar_maximise->img_over_surface = image_maximise;
				} else {
					window_maximised = TRUE;
					max_last_x = titlebar->x; max_last_y = titlebar->y;
					max_last_w = window->w; max_last_h = window->h;
					max_last_draggable = titlebar->draggable;
					titlebar->x = 0 - ENG_GetObject(REC_DEMO_01_WINDOW_LEFT)->w;
					titlebar->y = 0;
					window->w = demo_main->w + ENG_GetObject(REC_DEMO_01_WINDOW_LEFT)->w + ENG_GetObject(REC_DEMO_01_WINDOW_RIGHT)->w;
					window->h = demo_main->h;
					titlebar->draggable = FALSE;
					titlebar_maximise->img_surface = titlebar_maximise->img_over_surface = image_restore;
				}
				demo_01_expand_everything();
				ASCIIText_ebook.img_w = textarea->w;
				ASCIIText_ebook.img_h = textarea->h;
				/* Refresh textarea and related objects */
				demo_01_init_line_array(&line_count, &line_index, &line_array[0], current_char, &ASCIIText_ebook);
				ENG_WriteASCIIText(&ASCIIText_ebook, ENG_TEXT_WRITE_PAGE);
				demo_01_calculate_page_lines(&ASCIIText_ebook, &page_lines);
				demo_01_refresh_page_navigation_controls(&line_count, &line_index, &ASCIIText_line_stats, TRUE);
			} else if (event->object->id == REC_DEMO_01_TITLEBAR_MINIMISE) {
				/* Minimise the window */
				titlebar->visible = FALSE;
				minimised->visible = TRUE;
			} else if (event->object->id == LYR_DEMO_01_MINIMISED) {
				/* Restore the window */
				titlebar->visible = TRUE;
				minimised->visible = FALSE;
			} else if (event->object->id == REC_DEMO_01_TOOLBAR_FONT_SIZE) {
				/* Toggle font size */
				if (ASCIIText_ebook.font == font_ascii_16x24_white) {
					ASCIIText_ebook.font = font_ascii_8x12_white;
					ASCIIText_ebook.char_spacing = -1;
				} else if (ASCIIText_ebook.font == font_ascii_8x12_white) {
					ASCIIText_ebook.font = font_ascii_16x24_white;
					ASCIIText_ebook.char_spacing = -2;
				} else if (ASCIIText_ebook.font == font_ascii_16x24_black) {
					ASCIIText_ebook.font = font_ascii_8x12_black;
					ASCIIText_ebook.char_spacing = -1;
				} else if (ASCIIText_ebook.font == font_ascii_8x12_black) {
					ASCIIText_ebook.font = font_ascii_16x24_black;
					ASCIIText_ebook.char_spacing = -2;
				}
				/* Refresh textarea and related objects */
				demo_01_init_line_array(&line_count, &line_index, &line_array[0], current_char, &ASCIIText_ebook);
				ENG_WriteASCIIText(&ASCIIText_ebook, ENG_TEXT_WRITE_PAGE);
				demo_01_calculate_page_lines(&ASCIIText_ebook, &page_lines);
				demo_01_refresh_page_navigation_controls(&line_count, &line_index, &ASCIIText_line_stats, TRUE);
			} else if (event->object->id == REC_DEMO_01_TOOLBAR_FONT) {
				/* Toggle font and page colours */
				if (ASCIIText_ebook.font == font_ascii_16x24_white) {
					ASCIIText_ebook.font = font_ascii_16x24_black;
					ASCIIText_ebook.bgcolour = 0xffffff;
				} else if (ASCIIText_ebook.font == font_ascii_16x24_black) {
					ASCIIText_ebook.font = font_ascii_16x24_white;
					ASCIIText_ebook.bgcolour = 0x000000;
				} else if (ASCIIText_ebook.font == font_ascii_8x12_white) {
					ASCIIText_ebook.font = font_ascii_8x12_black;
					ASCIIText_ebook.bgcolour = 0xffffff;
				} else if (ASCIIText_ebook.font == font_ascii_8x12_black) {
					ASCIIText_ebook.font = font_ascii_8x12_white;
					ASCIIText_ebook.bgcolour = 0x000000;
				}
				ENG_WriteASCIIText(&ASCIIText_ebook, ENG_TEXT_WRITE_PAGE);
				scrollbar_up->bgcolour = scrollbar_down->bgcolour = ASCIIText_ebook.bgcolour;
			} else if (event->object->id == REC_DEMO_01_TOOLBAR_OPACITY) {
				/* Toggle opacities */
				if (titlebar->alpha == 128 && textarea->alpha == 255) {
					titlebar->alpha = titlebar->alpha_over = textarea->alpha = textarea->alpha_over = 128;
				} else if (titlebar->alpha == 128 && textarea->alpha == 128) {
					titlebar->alpha = titlebar->alpha_over = textarea->alpha = textarea->alpha_over = 255;
					scrollbar_up->bgcolour = scrollbar_down->bgcolour = ASCIIText_ebook.bgcolour;
				} else if (titlebar->alpha == 255 && textarea->alpha == 255) {
					titlebar->alpha = titlebar->alpha_over = 128;
					scrollbar_up->bgcolour = scrollbar_down->bgcolour = ASCIIText_ebook.bgcolour;
				}
			} else if (event->object->id == REC_DEMO_01_TOOLBAR_FULLSCREEN) {
				/* Cycle through different fullscreen modes */
				if (window_fullscreen == 0) {
					window_fullscreen = 1;
					fs_last_x = titlebar->x; fs_last_y = titlebar->y;
					fs_last_w = window->w; fs_last_h = window->h;
					fs_last_draggable = titlebar->draggable;
					titlebar->x = 0 - ENG_GetObject(REC_DEMO_01_WINDOW_LEFT)->w;
					titlebar->y = 0 - titlebar->h;
					window->w = demo_main->w + ENG_GetObject(REC_DEMO_01_WINDOW_LEFT)->w + ENG_GetObject(REC_DEMO_01_WINDOW_RIGHT)->w;
					window->h = demo_main->h + titlebar->h;
					titlebar->draggable = FALSE;
				} else if (window_fullscreen == 1) {
					window_fullscreen = 2;
					window->h += statusbar->h;
				} else {
					window_fullscreen = 0;
					titlebar->x = fs_last_x; titlebar->y = fs_last_y;
					window->w = fs_last_w; window->h = fs_last_h;
					titlebar->draggable = fs_last_draggable;
				}
				demo_01_expand_everything();
				ASCIIText_ebook.img_w = textarea->w;
				ASCIIText_ebook.img_h = textarea->h;
				/* Refresh textarea and related objects */
				/* It's not always necessary to initialise the line array here */
				if (window_maximised == FALSE && window_fullscreen < 2) {
					demo_01_init_line_array(&line_count, &line_index, &line_array[0], current_char, &ASCIIText_ebook);
				}
				ENG_WriteASCIIText(&ASCIIText_ebook, ENG_TEXT_WRITE_PAGE);
				demo_01_calculate_page_lines(&ASCIIText_ebook, &page_lines);
				demo_01_refresh_page_navigation_controls(&line_count, &line_index, &ASCIIText_line_stats, TRUE);
			}
			break;
		case ENG_EVENT_RELEASE:
			break;
		case ENG_EVENT_DRAG:
			if (event->object->id == REC_DEMO_01_SCROLLBAR_HANDLE) {
				/* Display the line of text the handle position represents */
				if ((scrollbar_pob->h - scrollbar_handle->h) == 0 || line_count - 1 == 0) {	/* Guard against divide by zero */
					line_index = 0;
				} else {
					line_index = scrollbar_handle->y / ((scrollbar_pob->h - scrollbar_handle->h) / ((float)line_count - 1));
				}
				ASCIIText_ebook.text = current_char = line_array[line_index];
				ENG_WriteASCIIText(&ASCIIText_ebook, ENG_TEXT_WRITE_PAGE);
				demo_01_refresh_page_navigation_controls(&line_count, &line_index, &ASCIIText_line_stats, FALSE);
			}
			break;
		case ENG_EVENT_QUITTING:
			/* Save demo_01.ebook_text_offset */
			sprintf(ebook_text_offset, "%i", current_char - demo_01_ebook_text);
			if (ENG_SetConfigOption("demo_01.ebook_text_offset", ebook_text_offset)) printf("%s: %s\n", __func__, ENG_GetError());
			break;
		case ENG_EVENT_POST_LOAD_OBJECTS:
			/* Store the objects_file so we know what to unload */
			strcpy(objects_file, ENG_GetObject(LYR_DEMO_01_TITLEBAR)->objects_file);
			/* Create LYR_DEMO_01_MAIN here because we want to attach an
			   image as near to the screen size as we have available */
			ENG_SetObjectDefaults(&new_object);
			new_object.id = LYR_DEMO_01_MAIN;
			new_object.klasse = ENG_OBJECT_KLASSE_LAYER;
			new_object.visible = TRUE;
			new_object.enabled = TRUE;
			new_object.alpha = 255;
			new_object.alpha_over = 255;
			new_object.w = op_env.screen_w;
			new_object.h = op_env.screen_h;
			if (op_env.screen_w < 640) {
				strcpy(new_object.img_file, PACKAGE_DATA_DIR "/images/demo_01/wallpaper-320x240.bmp");
			} else {
				strcpy(new_object.img_file, PACKAGE_DATA_DIR "/images/demo_01/wallpaper-640x480.bmp");
			}
			strcpy(new_object.img_over_file, new_object.img_file);
			if (ENG_CreateObject(&new_object, TRUE)) printf("%s: %s\n", __func__, ENG_GetError());
			/* Store pointers to commonly used objects */
			demo_main = ENG_GetObject(LYR_DEMO_01_MAIN);
			minimised = ENG_GetObject(LYR_DEMO_01_MINIMISED);
			window = ENG_GetObject(REC_DEMO_01_WINDOW);
			titlebar = ENG_GetObject(LYR_DEMO_01_TITLEBAR);
			titlebar_maximise = ENG_GetObject(REC_DEMO_01_TITLEBAR_MAXIMISE);
			textarea = ENG_GetObject(TXT_DEMO_01_TEXTAREA);
			scrollbar_pob = ENG_GetObject(REC_DEMO_01_SCROLLBAR_POB);
			scrollbar_up = ENG_GetObject(REC_DEMO_01_SCROLLBAR_UP);
			scrollbar_down = ENG_GetObject(REC_DEMO_01_SCROLLBAR_DOWN);
			scrollbar_handle = ENG_GetObject(REC_DEMO_01_SCROLLBAR_HANDLE);
			statusbar = ENG_GetObject(REC_DEMO_01_STATUSBAR);
			statusbar_line_stats = ENG_GetObject(TXT_DEMO_01_STATUSBAR_LINE_STATS);
			/* Set the zorder of the newly created LYR_DEMO_01_MAIN */
			ENG_SetObjectZOrder(demo_main, ENG_ZORDER_BELOW, minimised);
			/* Store pointers to the font image surfaces */
			font_ascii_8x12_white = ENG_GetObject(ENG_FNT_ASCII_8X12_WHITE)->img_surface;
			font_ascii_8x12_black = ENG_GetObject(ENG_FNT_ASCII_8X12_BLACK)->img_surface;
			font_ascii_8x12_white_black_outline = ENG_GetObject(ENG_FNT_ASCII_8X12_WHITE_BLACK_OUTLINE)->img_surface;
			font_ascii_16x24_white = ENG_GetObject(ENG_FNT_ASCII_16X24_WHITE)->img_surface;
			font_ascii_16x24_black = ENG_GetObject(ENG_FNT_ASCII_16X24_BLACK)->img_surface;
			font_ascii_16x24_white_black_outline = ENG_GetObject(ENG_FNT_ASCII_16X24_WHITE_BLACK_OUTLINE)->img_surface;
			/* Store pointers to the maximise and restore image surfaces */
			image_maximise = titlebar_maximise->img_surface;
			image_restore = titlebar_maximise->img_over_surface;
			titlebar_maximise->img_over_surface = titlebar_maximise->img_surface;
			/* Initialise the minimised window size */
			minimised->y = demo_main->h - minimised->h;
			minimised->w = demo_main->w / 4;
			/* Set a default window size */
			window->w = demo_main->w - demo_main->w / 4;
			window->h = demo_main->h - demo_main->h / 4;
			titlebar->x = (demo_main->w - window->w) / 2;
			titlebar->y = (demo_main->h - window->h) / 2;
			window_fullscreen = window_maximised = 0;
			demo_01_expand_everything();
			fs_last_x = max_last_x = titlebar->x;
			fs_last_y = max_last_y = titlebar->y;
			fs_last_w = max_last_w = window->w;
			fs_last_h = max_last_h = window->h;
			fs_last_draggable = max_last_draggable = titlebar->draggable;
			/* Load the ebook before we reference demo_01_ebook_text below */
			ebook_loaded = demo_01_load_ebook();
			/* Initialise the ASCIIText_ebook object */
			ASCIIText_ebook.font = font_ascii_8x12_black;
			ASCIIText_ebook.target = textarea->img_surface;
			ASCIIText_ebook.img_x = textarea->img_x;
			ASCIIText_ebook.img_y = textarea->img_y;
			ASCIIText_ebook.img_w = textarea->img_w;
			ASCIIText_ebook.img_h = textarea->img_h;
			ASCIIText_ebook.bgcolour = 0xffffff;
			ASCIIText_ebook.char_spacing = -1;
			ASCIIText_ebook.line_spacing = 0;
			ASCIIText_ebook.tab_spaces = 8;
			ASCIIText_ebook.fixed_width = FALSE;
			ASCIIText_ebook.render = TRUE;
			ASCIIText_ebook.text = current_char = demo_01_ebook_text;
			/* Initialise the ASCIIText_line_stats object */
			ASCIIText_line_stats.font = font_ascii_8x12_white_black_outline;
			ASCIIText_line_stats.target = statusbar_line_stats->img_surface;
			ASCIIText_line_stats.img_x = statusbar_line_stats->img_x;
			ASCIIText_line_stats.img_y = statusbar_line_stats->img_y;
			ASCIIText_line_stats.img_w = statusbar_line_stats->img_w;
			ASCIIText_line_stats.img_h = statusbar_line_stats->img_h;
			ASCIIText_line_stats.bgcolour = (ENG_COLOUR_KEY_RED << 16) | (ENG_COLOUR_KEY_GREEN << 8) | (ENG_COLOUR_KEY_BLUE);
			ASCIIText_line_stats.char_spacing = -1;
			ASCIIText_line_stats.line_spacing = 0;
			ASCIIText_line_stats.tab_spaces = 8;
			ASCIIText_line_stats.fixed_width = FALSE;
			ASCIIText_line_stats.render = TRUE;
			ASCIIText_line_stats.text = demo_01_line_stats_text;
			/* Retrieve demo_01.ebook_text_offset if possible */
			if (ENG_GetConfigOption("demo_01.ebook_text_offset", ebook_text_offset)) {
				printf("%s: %s\n", __func__, ENG_GetError());
			} else {
				if ((sscanf(ebook_text_offset, "%i", &offset)) == 1) current_char += offset;
			}
			/* Write the first page of the ebook */
			demo_01_init_line_array(&line_count, &line_index, &line_array[0], current_char, &ASCIIText_ebook);
			ENG_WriteASCIIText(&ASCIIText_ebook, ENG_TEXT_WRITE_PAGE);
			demo_01_calculate_page_lines(&ASCIIText_ebook, &page_lines);
			demo_01_refresh_page_navigation_controls(&line_count, &line_index, &ASCIIText_line_stats, TRUE);
		case ENG_EVENT_SELECT:
			if (event->object->id == REC_DEMO_01_SCROLLBAR_BUTTON_DOWN) {
				/* Display the next line of the ebook */
				if (line_index < line_count - 1) {
					line_index++;
					ASCIIText_ebook.text = current_char = line_array[line_index];
					ENG_WriteASCIIText(&ASCIIText_ebook, ENG_TEXT_WRITE_PAGE);
					demo_01_refresh_page_navigation_controls(&line_count, &line_index, &ASCIIText_line_stats, TRUE);
				}
			} else if (event->object->id == REC_DEMO_01_SCROLLBAR_BUTTON_UP) {
				/* Display the previous line of the ebook */
				if (line_index > 0) {
					line_index--;
					ASCIIText_ebook.text = current_char = line_array[line_index];
					ENG_WriteASCIIText(&ASCIIText_ebook, ENG_TEXT_WRITE_PAGE);
					demo_01_refresh_page_navigation_controls(&line_count, &line_index, &ASCIIText_line_stats, TRUE);
				}
			} else if (event->object->id == REC_DEMO_01_SCROLLBAR_DOWN) {
				/* Display the next n lines of the ebook */
				if (line_index < line_count - page_lines) {
					line_index += page_lines;
				} else {
					line_index = line_count - 1;
				}
				ASCIIText_ebook.text = current_char = line_array[line_index];
				ENG_WriteASCIIText(&ASCIIText_ebook, ENG_TEXT_WRITE_PAGE);
				demo_01_refresh_page_navigation_controls(&line_count, &line_index, &ASCIIText_line_stats, TRUE);
			} else if (event->object->id == REC_DEMO_01_SCROLLBAR_UP) {
				/* Display the previous n lines of the ebook */
				if (line_index >= page_lines) {
					line_index -= page_lines;
				} else {
					line_index = 0;
				}
				ASCIIText_ebook.text = current_char = line_array[line_index];
				ENG_WriteASCIIText(&ASCIIText_ebook, ENG_TEXT_WRITE_PAGE);
				demo_01_refresh_page_navigation_controls(&line_count, &line_index, &ASCIIText_line_stats, TRUE);
			}
			break;
		case ENG_EVENT_SHUTDOWN:
			ENG_Quit();
			break;
		case ENG_EVENT_TOGGLE_FULLSCREEN:
			op_env.fullscreen = 1 - op_env.fullscreen;
			if (ENG_SetOpEnvFullScreen(op_env.fullscreen)) printf("%s: %s\n", __func__, ENG_GetError());
			break;
	}
}

/***************************************************************************
 * Demo 01 - Load ebook                                                    *
 ***************************************************************************/
/* This loads the ebook from disk */

int demo_01_load_ebook(void) {
	FILE *fp;
	long file_size;
	int retval = FALSE;

	/* Open the ebook */
	if ((fp = fopen(PACKAGE_DATA_DIR "/ebooks/ebook.txt", "r")) == NULL) {
		sprintf(demo_01_ebook_error_text, "Error loading ebook :-\n\nCouldn't open %s\n", PACKAGE_DATA_DIR "/ebooks/ebook.txt");
		demo_01_ebook_text = demo_01_ebook_error_text;
	} else {
		/* Get file size */
		fseek(fp, 0, SEEK_END);
		file_size = ftell(fp);
		rewind(fp);
		/* Allocate enough memory for the ebook and a terminating zero */
		if ((demo_01_ebook_text = (char*) malloc(sizeof(char) * (file_size + 1))) == NULL) {
			sprintf(demo_01_ebook_error_text, "Error loading ebook :-\n\nCouldn't allocate %li bytes for %s\n", file_size + 1, PACKAGE_DATA_DIR "/ebooks/ebook.txt");
			demo_01_ebook_text = demo_01_ebook_error_text;
		} else {
			retval = TRUE;
			fread(demo_01_ebook_text, 1, file_size, fp);
		}			
		fclose(fp);
	}
	
	return retval;
}

/***************************************************************************
 * Demo 01 - Expand Everything                                             *
 ***************************************************************************/
/* This expands everything to fit the underlying unpainted window object */

void demo_01_expand_everything(void) {
	ENG_Object *window = ENG_GetObject(REC_DEMO_01_WINDOW);
	ENG_Object *titlebar = ENG_GetObject(LYR_DEMO_01_TITLEBAR);
	ENG_Object *titlebar_exit = ENG_GetObject(REC_DEMO_01_TITLEBAR_EXIT);
	ENG_Object *titlebar_maximise = ENG_GetObject(REC_DEMO_01_TITLEBAR_MAXIMISE);
	ENG_Object *titlebar_minimise = ENG_GetObject(REC_DEMO_01_TITLEBAR_MINIMISE);
	ENG_Object *toolbar = ENG_GetObject(REC_DEMO_01_TOOLBAR);
	ENG_Object *window_left = ENG_GetObject(REC_DEMO_01_WINDOW_LEFT);
	ENG_Object *window_right = ENG_GetObject(REC_DEMO_01_WINDOW_RIGHT);
	ENG_Object *statusbar = ENG_GetObject(REC_DEMO_01_STATUSBAR);
	ENG_Object *textarea = ENG_GetObject(TXT_DEMO_01_TEXTAREA);
	ENG_Object *scrollbar_pob = ENG_GetObject(REC_DEMO_01_SCROLLBAR_POB);
	ENG_Object *scrollbar_button_up = ENG_GetObject(REC_DEMO_01_SCROLLBAR_BUTTON_UP);
	ENG_Object *scrollbar_button_down = ENG_GetObject(REC_DEMO_01_SCROLLBAR_BUTTON_DOWN);
	ENG_Object *shadow_e = ENG_GetObject(REC_DEMO_01_SHADOW_E);
	ENG_Object *shadow_s = ENG_GetObject(REC_DEMO_01_SHADOW_S);

	if (window->w < 136) window->w = 136; if (window->h < 152) window->h = 152;

	titlebar->w = toolbar->w = statusbar->w = window->w;
	titlebar_exit->x = window->w - window_right->w - titlebar_exit->w;
	titlebar_maximise->x = titlebar_exit->x - 1 - titlebar_maximise->w;
	titlebar_minimise->x = titlebar_maximise->x - 1 - titlebar_minimise->w;
	window_left->h = window_right->h = window->h - titlebar->h - toolbar->h - statusbar->h;
	window_right->x = window->w - window_right->w;
	scrollbar_pob->x = window->w - window_right->w - scrollbar_pob->w;
	scrollbar_pob->h = window->h - titlebar->h - toolbar->h - statusbar->h - scrollbar_button_up->h - scrollbar_button_down->h;
	scrollbar_button_down->y = scrollbar_pob->h;
	textarea->w = textarea->img_w = textarea->img_over_w = window->w - window_left->w - window_right->w - scrollbar_pob->w;
	textarea->h = textarea->img_h = textarea->img_over_h = window->h - titlebar->h - toolbar->h - statusbar->h;
	statusbar->y = window->h - statusbar->h;
	shadow_e->x = window->w;
	shadow_e->h = window->h;
	shadow_s->y = window->h;
	shadow_s->w = window->w - shadow_s->x;
}

/***************************************************************************
 * Demo 01 - Initialise Line Array                                         *
 ***************************************************************************/
/* This stores pointers to each line of text, sets line_count and
   line_index based upon the current_char */

void demo_01_init_line_array(int *line_count, int *line_index, char *line_array[], char *current_char, ENG_ASCIIText *ASCIIText_ebook) {

	/* Index the ebook by the current width of the surface region */
	ASCIIText_ebook->render = FALSE;
	ASCIIText_ebook->text = demo_01_ebook_text;
	*line_count = 0;
	do {
		line_array[*line_count] = ASCIIText_ebook->text;
		*line_count += 1;
		ENG_WriteASCIIText(ASCIIText_ebook, ENG_TEXT_WRITE_LINE);
		ASCIIText_ebook->text = ASCIIText_ebook->next_char;
	} while (*ASCIIText_ebook->text != 0);
	ASCIIText_ebook->text = demo_01_ebook_text;
	ASCIIText_ebook->render = TRUE;
	/* Set line_index based upon current_char */
	*line_index = 0;
	do {
		if (line_array[*line_index] >= current_char) {
			if (line_array[*line_index] > current_char) *line_index -= 1;
			ASCIIText_ebook->text = line_array[*line_index];
			break;
		}
		*line_index += 1;
	} while (*line_index < *line_count);
}

/***************************************************************************
 * Demo 01 - Calculate Page Lines                                          *
 ***************************************************************************/
/* This calculates the maximum number of lines the current size of the
   textarea can accommodate and is used for scrolling by page */

void demo_01_calculate_page_lines(ENG_ASCIIText *ASCIIText_ebook, int *page_lines) {

	*page_lines = ASCIIText_ebook->img_h / (ASCIIText_ebook->font->h - 1 + ASCIIText_ebook->line_spacing);
}

/***************************************************************************
 * Demo 01 - Refresh Page Navigation Controls                              *
 ***************************************************************************/
/* This maintains the objects/controls that need to be in sync with the
   current state of the ebook */

void demo_01_refresh_page_navigation_controls(int *line_count, int *line_index, ENG_ASCIIText *ASCIIText_line_stats, int adjust_handle) {
	ENG_Object *scrollbar_pob = ENG_GetObject(REC_DEMO_01_SCROLLBAR_POB);
	ENG_Object *scrollbar_handle = ENG_GetObject(REC_DEMO_01_SCROLLBAR_HANDLE);
	ENG_Object *scrollbar_up = ENG_GetObject(REC_DEMO_01_SCROLLBAR_UP);
	ENG_Object *scrollbar_down = ENG_GetObject(REC_DEMO_01_SCROLLBAR_DOWN);

	if (adjust_handle) {
		/* This positions the scrollbar handle to reflect line_index */
		if (*line_count - 1 == 0) {	/* Guard against divide by zero */
			scrollbar_handle->y = 0;
		} else {
			scrollbar_handle->y = *line_index * ((scrollbar_pob->h - scrollbar_handle->h) / ((float)*line_count - 1));
		}
	}
	/* This resizes the scrollbar up and down regions */
	scrollbar_up->h = scrollbar_handle->y;
	scrollbar_down->y = scrollbar_handle->y + scrollbar_handle->h;
	scrollbar_down->h = scrollbar_pob->h - scrollbar_down->y;
	/* This writes the statusbar line stats */
	sprintf(demo_01_line_stats_text, "Line %i of %i", *line_index + 1, *line_count);
	ENG_WriteASCIIText(ASCIIText_line_stats, ENG_TEXT_WRITE_PAGE);
}

















