/* Tetravex (c) Thunor 2008 <thunorsif@hotmail.com>
www.thunor.org.uk

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. */


/* Includes */
#include <SDL/SDL_main.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "main.h"


/* Defines */
/* Main key object ids */
#define LYR_MAIN 100
#define REC_MAIN_EXIT 101
#define REC_MAIN_TETRAVEX 102
#define REC_MAIN_TETRAVEX_LOGO 103
#define REC_MAIN_TETRAVEX_MOVE 104


/***************************************************************************
 * Main                                                                    *
 ***************************************************************************/

#ifdef __cplusplus
extern "C"
#endif
int main(int argc, char *argv[]) {
	int retval = ENG_ERROR_NONE;

	strcpy(home, getenv("HOME"));

	/* Set rcfile */
	#if defined(PLATFORM_PC)
		/*strcpy(rcfile, home);
		strcat(rcfile, "/.engrc");*/
		strcpy(rcfile, "./engrc");
	#elif defined(PLATFORM_ZAURUS)
		strcpy(rcfile, home);
		strcat(rcfile, "/.engrc");
	#elif defined(PLATFORM_GP2X)
		strcpy(rcfile, "./engrc");
	#endif
	
	/* Set engine operating environment */
	#if defined(PLATFORM_PC)
		/*strcpy(op_env.image_dir, PACKAGE_DATA_DIR "/images/main/320x240");
		strcpy(op_env.objects_file, PACKAGE_DATA_DIR "/objects/main-320x240");
		strcpy(op_env.config_file, rcfile);
		op_env.screen_w = 320;
		op_env.screen_h = 240;
		op_env.canvas_w = 320;
		op_env.canvas_h = 240;*/
		strcpy(op_env.image_dir, PACKAGE_DATA_DIR "/images/main/640x480");
		strcpy(op_env.objects_file, PACKAGE_DATA_DIR "/objects/main-640x480");
		strcpy(op_env.config_file, rcfile);
		op_env.screen_w = 640;
		op_env.screen_h = 480;
		op_env.canvas_w = 640;
		op_env.canvas_h = 480;
		strcpy(op_env.title, "Tetravex");
		op_env.fullscreen = FALSE;
		op_env.systempointer = TRUE;//FALSE;
		op_env.pointer = TRUE;//FALSE
		op_env.pointer_mode = ENG_POINTER_MODE_SIMULATED;
		op_env.pointer_move_unit_low = 8;
		op_env.pointer_move_unit_high = 16;
		op_env.pointer_move_unit_threshold = 12;
	#elif defined(PLATFORM_ZAURUS)
		strcpy(op_env.image_dir, PACKAGE_DATA_DIR "/images/main/320x240");
		strcpy(op_env.objects_file, PACKAGE_DATA_DIR "/objects/main-320x240");
		strcpy(op_env.config_file, rcfile);
		op_env.screen_w = 320;
		op_env.screen_h = 240;
		op_env.canvas_w = 320;
		op_env.canvas_h = 240;
		strcpy(op_env.title, "Tetravex");
		op_env.fullscreen = TRUE;
		op_env.systempointer = FALSE;
		op_env.pointer = TRUE;
		op_env.pointer_mode = ENG_POINTER_MODE_SIMULATED;//ENG_POINTER_MODE_PHYSICAL;
		op_env.pointer_move_unit_low = 4;
		op_env.pointer_move_unit_high = 8;
		op_env.pointer_move_unit_threshold = 6;
	#elif defined(PLATFORM_GP2X)
		strcpy(op_env.image_dir, PACKAGE_DATA_DIR "/images/main/320x240");
		strcpy(op_env.objects_file, PACKAGE_DATA_DIR "/objects/main-320x240");
		strcpy(op_env.config_file, rcfile);
		op_env.screen_w = 320;
		op_env.screen_h = 240;
		op_env.canvas_w = 320;
		op_env.canvas_h = 240;
		strcpy(op_env.title, "Tetravex");
		op_env.fullscreen = TRUE;
		op_env.systempointer = FALSE;
		op_env.pointer = TRUE;
		op_env.pointer_mode = ENG_POINTER_MODE_SIMULATED;
		op_env.pointer_move_unit_low = 4;
		op_env.pointer_move_unit_high = 8;
		op_env.pointer_move_unit_threshold = 6;
	#endif
	op_env.delay = 0;
	op_env.cycle_limit = 50;
	op_env.viewport_x = 0;
	op_env.viewport_y = 0;
	op_env.viewport_w = op_env.screen_w;// - 100;
	op_env.viewport_h = op_env.screen_h;// - 100;
	op_env.viewport_screen_x = 0;
	op_env.viewport_screen_y = 0;
	op_env.event_processor = &main_event_processor;

	/* Initialise the engine (the config file is loaded if found) */
	retval = ENG_Init(&op_env);
	if (retval) printf("%s: %s\n", __func__, ENG_GetError());

	/* UPDATE OP_ENV BASED UPON CONFIG FILE OPTIONS HERE */
	/* UPDATE OP_ENV BASED UPON CONFIG FILE OPTIONS HERE */
	/* UPDATE OP_ENV BASED UPON CONFIG FILE OPTIONS HERE */

	/* UPDATE OP_ENV BASED UPON COMMAND LINE OPTIONS HERE */
	/* UPDATE OP_ENV BASED UPON COMMAND LINE OPTIONS HERE */
	/* UPDATE OP_ENV BASED UPON COMMAND LINE OPTIONS HERE */

	/* Start the engine (ENG_Quit stops it) */
	retval = ENG_Start();
	if (retval) printf("%s: %s\n", __func__, ENG_GetError());
		
	return retval;
}

/***************************************************************************
 * Main Event Processor                                                    *
 ***************************************************************************/

void main_event_processor(ENG_Event *event) {
	static int init_first = TRUE, init_on_return = FALSE, init_all = TRUE;
	static char last_game_played[ENG_CONFIG_OPTION_VALUE_LENGTH];
	ENG_Effect tetravex_effect, logo_effect;

	/* ENG_ShowEvent(event); */
	
	/* Initialise some things on the first run */
	if (init_first) {
		init_first = FALSE;
	}

	/* Initialise some things on return runs */
	if (init_on_return) {
		init_on_return = FALSE;
	}

	/* Initialise some things on all runs */
	if (init_all) {
		init_all = FALSE;
		ENG_GetObject(LYR_MAIN)->visible = TRUE;
		ENG_GetObject(REC_MAIN_TETRAVEX)->visible = FALSE;
		ENG_GetObject(REC_MAIN_TETRAVEX_LOGO)->visible = ENG_INHERITED;
		ENG_GetObject(REC_MAIN_TETRAVEX_MOVE)->visible = ENG_INHERITED;
		/* Fade-in REC_MAIN_TETRAVEX_LOGO */
		logo_effect.effect_id = LYR_MAIN;
		logo_effect.klasse = ENG_EFFECT_ALPHA_FADE;
		logo_effect.object = ENG_GetObject(REC_MAIN_TETRAVEX_LOGO);
		logo_effect.frames = op_env.cycle_limit * 0.25;
		logo_effect.paused = FALSE;
		logo_effect.src_alpha = logo_effect.src_alpha_over = 0;
		logo_effect.dst_alpha = logo_effect.dst_alpha_over = 255;
		if (ENG_CreateEffect(&logo_effect)) printf("%s: %s\n", __func__, ENG_GetError());
		/* Move REC_MAIN_TETRAVEX_MOVE in from the right */
		tetravex_effect.effect_id = LYR_MAIN;
		tetravex_effect.klasse = ENG_EFFECT_LINEAR_MOVE;
		tetravex_effect.object = ENG_GetObject(REC_MAIN_TETRAVEX_MOVE);
		tetravex_effect.frames = op_env.cycle_limit * 0.25;
		tetravex_effect.paused = FALSE;
		tetravex_effect.src_x = op_env.screen_w;
		tetravex_effect.src_y = tetravex_effect.dst_y = tetravex_effect.object->y;
		tetravex_effect.dst_x = tetravex_effect.object->x;
		if (ENG_CreateEffect(&tetravex_effect)) printf("%s: %s\n", __func__, ENG_GetError());
	}

	/* Process events */
	switch (event->klasse) {
		case ENG_EVENT_CLICK:
			if (event->object->id == REC_MAIN_EXIT) {
				ENG_Quit();
			} else if (event->object->id == REC_MAIN_TETRAVEX) {
				/* Retrieve tetravex.last_game_played if possible */
				if (ENG_GetConfigOption("tetravex.last_game_played", last_game_played)) printf("%s: %s\n", __func__, ENG_GetError());
				/* Load the 3x3 or 4x4 game */
				if (!strcmp(last_game_played, "4x4")) {
					/* Set the event processor to tetravex_4x4 */
					op_env.event_processor = &tetravex_4x4_event_processor;
					/* Set the tetravex_4x4 image directory */
					/* Set the tetravex_4x4 objects file */
					if (op_env.screen_w < 640) {
						strcpy(op_env.image_dir, PACKAGE_DATA_DIR "/images/tetravex/320x240");
						strcpy(op_env.objects_file, PACKAGE_DATA_DIR "/objects/tetravex-4x4-320x240");
					} else {
						strcpy(op_env.image_dir, PACKAGE_DATA_DIR "/images/tetravex/640x480");
						strcpy(op_env.objects_file, PACKAGE_DATA_DIR "/objects/tetravex-4x4-640x480");
					}
				} else {
					/* Set the event processor to tetravex_3x3 */
					op_env.event_processor = &tetravex_3x3_event_processor;
					/* Set the tetravex_3x3 image directory */
					/* Set the tetravex_3x3 objects file */
					if (op_env.screen_w < 640) {
						strcpy(op_env.image_dir, PACKAGE_DATA_DIR "/images/tetravex/320x240");
						strcpy(op_env.objects_file, PACKAGE_DATA_DIR "/objects/tetravex-3x3-320x240");
					} else {
						strcpy(op_env.image_dir, PACKAGE_DATA_DIR "/images/tetravex/640x480");
						strcpy(op_env.objects_file, PACKAGE_DATA_DIR "/objects/tetravex-3x3-640x480");
					}
				}
				if (ENG_SetOpEnvEventProcessor(op_env.event_processor)) printf("%s: %s\n", __func__, ENG_GetError());
				if (ENG_SetOpEnvImageDir(op_env.image_dir)) printf("%s: %s\n", __func__, ENG_GetError());
				if (ENG_SetOpEnvObjectsFile(op_env.objects_file)) printf("%s: %s\n", __func__, ENG_GetError());
				if (ENG_LoadObjects(TRUE)) printf("%s: %s\n", __func__, ENG_GetError());
				/* Hide the current layer */
				ENG_GetObject(LYR_MAIN)->visible = FALSE;
				init_all = TRUE;
			}
			break;
		case ENG_EVENT_QUITTING:
			break;
		case ENG_EVENT_POST_EFFECT:
			if (event->object->id == REC_MAIN_TETRAVEX_LOGO) {
				event->object->visible = FALSE;
			} else if (event->object->id == REC_MAIN_TETRAVEX_MOVE) {
				event->object->visible = FALSE;
				ENG_GetObject(REC_MAIN_TETRAVEX)->visible = ENG_INHERITED;
			}
			break;
		case ENG_EVENT_SHUTDOWN:
			ENG_Quit();
			break;
		case ENG_EVENT_TOGGLE_FULLSCREEN:
			op_env.fullscreen = 1 - op_env.fullscreen;
			if (ENG_SetOpEnvFullScreen(op_env.fullscreen)) printf("%s: %s\n", __func__, ENG_GetError());
			break;
	}

}







