/* Tetravex (c) Thunor 2008 <thunorsif@hotmail.com>
www.thunor.org.uk

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. */


/* Includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "main.h"


/* Defines */
/* Tetravex key object ids */
#define LYR_TETRAVEX 200
#define REC_TETRAVEX_OPTIONS 201
#define REC_TETRAVEX_HELP 202
#define REC_TETRAVEX_EXIT 203
#define REC_TETRAVEX_3X3 204
#define REC_TETRAVEX_4X4 205
#define REC_TETRAVEX_HISCORES 206
#define REC_TETRAVEX_SOLVE 207
#define REC_TETRAVEX_TIMER_1S 209
#define REC_TETRAVEX_TIMER_10S 210
#define REC_TETRAVEX_TIMER_1M 211
#define REC_TETRAVEX_TIMER_10M 212
#define REC_TETRAVEX_SLOTS_00 300
#define LYR_TETRAVEX_TILES_00N 400
#define REC_TETRAVEX_TILE_SHADOW_E 500

/* Timer states */
#define TIMER_OFF 0
#define TIMER_ON 1
#define TIMER_PAUSED 2
#define TIMER_MAX 59 * 60 + 59

#define HISCORES_MAX 5


/***************************************************************************
 * Tetravex 3x3 Event Processor                                            *
 ***************************************************************************/

void tetravex_3x3_event_processor(ENG_Event *event) {
	static int init_on_return = FALSE;
	static SDL_Surface *tetravex_tile_images[10];
	static int tilelocations[18];	/* The index is the slot id adjusted, the contents the north tile id */
	static ENG_Object *shadoweast, *shadowsouth, *timer1s, *timer10s, *timer1m, *timer10m;
	static int timer_state, timer_lastticks, timer_time, solved = FALSE, hiscore_time[HISCORES_MAX];
	static char objects_file[ENG_PATH_LENGTH];
	int count, count2, x, y, randval, swap, match;
	char name[ENG_CONFIG_OPTION_NAME_LENGTH], value[ENG_CONFIG_OPTION_VALUE_LENGTH];
	ENG_Object *object, *target;
	ENG_Event myevent;
	ENG_Effect effect;
	
	/* Initialise some things on return runs */
	if (init_on_return) {
		init_on_return = FALSE;
		ENG_GetObject(LYR_TETRAVEX)->visible = TRUE;
		/* Unpause the timer if paused */
		if (timer_state == TIMER_PAUSED) {
			timer_state = TIMER_ON;
			timer_lastticks = SDL_GetTicks();
		}
	}

	/* Process events */
	switch (event->klasse) {
		case ENG_EVENT_CYCLE:
			/* This is the game timer which increments in units of 1 second up to
			   3599 seconds or 59:59 at which point it turns itself off */
			if (timer_state == TIMER_ON) {
				if (SDL_GetTicks() - timer_lastticks >= 1000) {
					if (timer_time < TIMER_MAX) {
						timer_lastticks = SDL_GetTicks();
						timer_time++;
					} else {
						timer_state = TIMER_OFF;
					}
					/* Update the timer visuals. The img_over_x properties need not be set
					   because the objects are disabled and hence don't generate over events */
					timer1s->img_x = timer_time % 10 * timer1s->w;
					timer10s->img_x = timer_time % 60 / 10 % 10 * timer1s->w;
					timer1m->img_x = timer_time / 60 % 10 * timer1s->w;
					timer10m->img_x = timer_time / 60 / 10 % 10 * timer1s->w;
				}
			}
			break;
		case ENG_EVENT_OVER:
			/* temp temp
			printf("over: id=%i\n", event->object->id); */
			break;
		case ENG_EVENT_OUT:
			/* temp temp
			printf("out: id=%i\n", event->object->id); */
			break;
		case ENG_EVENT_PRESS:
			if (event->object->id >= LYR_TETRAVEX_TILES_00N && event->object->id <= LYR_TETRAVEX_TILES_00N + 8 * 4) {
				/* Manage the zorder of the tiles and shadows */
				ENG_SetObjectZOrder(event->object, ENG_ZORDER_ABOVE_ALL, event->object->pob);				/* North */
				ENG_SetObjectZOrder(ENG_GetObject(event->object->id + 1), ENG_ZORDER_BELOW, event->object);	/* East */
				ENG_SetObjectZOrder(ENG_GetObject(event->object->id + 2), ENG_ZORDER_BELOW, event->object);	/* South */
				ENG_SetObjectZOrder(ENG_GetObject(event->object->id + 3), ENG_ZORDER_BELOW, event->object);	/* West */
				/* Attach the tile shadows to the east tile and make them visible */
				target = ENG_GetObject(event->object->id + 1);	/* East tile */
				ENG_SetObjectZOrder(shadoweast, ENG_ZORDER_BELOW, target);
				shadoweast->pob = target; shadoweast->visible = TRUE;	
				target = ENG_GetObject(event->object->id + 1);
				ENG_SetObjectZOrder(shadowsouth, ENG_ZORDER_BELOW, target);
				shadowsouth->pob = target; shadowsouth->visible = TRUE;
			}
			break;
		case ENG_EVENT_CLICK:
			/* Finalise all effects now as objects may get freed */
			ENG_ControlEffects(LYR_TETRAVEX, ENG_EFFECT_STATE_FINALISE);

			if (event->object->id == REC_TETRAVEX_OPTIONS) {
 				/* vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv temp temp will be loaded from Options layer */
				/* Change the event processor to demo_01 */
				op_env.event_processor = &demo_01_event_processor;
				if (ENG_SetOpEnvEventProcessor(op_env.event_processor)) printf("%s: %s\n", __func__, ENG_GetError());
				/* Set the demo_01 image directory */
				strcpy(op_env.image_dir, PACKAGE_DATA_DIR "/images/demo_01");
				if (ENG_SetOpEnvImageDir(op_env.image_dir)) printf("%s: %s\n", __func__, ENG_GetError());
				/* Load the demo_01 objects file */
				strcpy(op_env.objects_file, PACKAGE_DATA_DIR "/objects/demo_01");
				if (ENG_SetOpEnvObjectsFile(op_env.objects_file)) printf("%s: %s\n", __func__, ENG_GetError());
				if (ENG_LoadObjects(TRUE)) printf("%s: %s\n", __func__, ENG_GetError());
				/* Hide the current layer */
				ENG_GetObject(LYR_TETRAVEX)->visible = FALSE;
				/* Pause the timer if on */
				if (timer_state == TIMER_ON) timer_state = TIMER_PAUSED;
				init_on_return = TRUE;
				/* ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ temp temp will be loaded from Options layer */
			} else if (event->object->id == REC_TETRAVEX_HELP) {
				/* temp temp */
				object = ENG_GetObject(ENG_LYR_SCREEN);
				target = ENG_GetObject(ENG_LYR_VIEWPORT);
				if (object->bgcolour == 0x0000cc) {
					object->bgcolour = object->bgcolour_over = ENG_UNDEFINED;
					target->bgcolour = target->bgcolour_over = ENG_UNDEFINED;
				} else {
					object->bgcolour = object->bgcolour_over = 0x0000cc;
					target->bgcolour = target->bgcolour_over = 0xcc0000;
				}
			} else if (event->object->id == REC_TETRAVEX_EXIT) {
				/* Change the event processor to main */
				op_env.event_processor = &main_event_processor;
				if (ENG_SetOpEnvEventProcessor(op_env.event_processor)) printf("%s: %s\n", __func__, ENG_GetError());
				/* Unload the tetravex_3x3 objects file */
				if (ENG_UnloadObjects(objects_file)) printf("%s: %s\n", __func__, ENG_GetError());
			} else if (event->object->id == REC_TETRAVEX_3X3) {
				/* Start a new 3x3 game */
				for (count = 0; count < 9; count++) {
					/* Reposition the tiles in the home slots */
					object = ENG_GetObject(LYR_TETRAVEX_TILES_00N + count * 4);
					target = ENG_GetObject(REC_TETRAVEX_SLOTS_00 + count);
					object->x = target->x; object->y = target->y;	/* East, south and west are attached to north */
					object->enabled = TRUE;	/* Reenable the north tiles */
					/* Randomise the tile images */
					for (count2 = 0; count2 < 4; count2++) {
						object = ENG_GetObject(LYR_TETRAVEX_TILES_00N + count * 4 + count2);
						object->img_surface = object->img_over_surface = tetravex_tile_images[rand() % 10];
					}
					/* Reset the tile slot locations */
					tilelocations[count] = LYR_TETRAVEX_TILES_00N + count * 4;
					tilelocations[count + 9] = ENG_UNDEFINED;
				}
				/* Set the tile adjoining edge images */
				for (count = 0; count < 9; count++) {
					if (count != 2 && count != 5 && count != 8) {	/* East = west */
						object = ENG_GetObject(LYR_TETRAVEX_TILES_00N + count * 4 + 1);
						object->img_surface = object->img_over_surface = ENG_GetObject(LYR_TETRAVEX_TILES_00N + (count + 1) * 4 + 3)->img_surface;
					}
					if (count < 6) {	/* South = north */
						object = ENG_GetObject(LYR_TETRAVEX_TILES_00N + count * 4 + 2);
						object->img_surface = object->img_over_surface = ENG_GetObject(LYR_TETRAVEX_TILES_00N + (count + 3) * 4)->img_surface;
					}
				}
				/* Shuffle the tile positions within the home slots */
				for (count = 0; count < 9; count++) {
					randval = (rand() % 9);
					object = ENG_GetObject(tilelocations[count]);
					target = ENG_GetObject(tilelocations[randval]);
					x = object->x; y = object->y;
					object->x = target->x; object->y = target->y;
					target->x = x; target->y = y;
					swap = tilelocations[count];
					tilelocations[count] = tilelocations[randval];
					tilelocations[randval] = swap;
				}
				/* Restart the timer */
				timer_state = TIMER_ON;
				timer_lastticks = 0;
				timer_time = -1;
				solved = FALSE;
			} else if (event->object->id == REC_TETRAVEX_4X4) {
				/* Change the event processor to tetravex_4x4 */
				op_env.event_processor = &tetravex_4x4_event_processor;
				if (ENG_SetOpEnvEventProcessor(op_env.event_processor)) printf("%s: %s\n", __func__, ENG_GetError());
				/* Unload the tetravex_3x3 objects file */
				if (ENG_UnloadObjects(objects_file)) printf("%s: %s\n", __func__, ENG_GetError());
				/* Set the tetravex_4x4 image directory */
				/* Load the tetravex_4x4 objects file */
				if (op_env.screen_w < 640) {
					strcpy(op_env.image_dir, PACKAGE_DATA_DIR "/images/tetravex/320x240");
					strcpy(op_env.objects_file, PACKAGE_DATA_DIR "/objects/tetravex-4x4-320x240");
				} else {
					strcpy(op_env.image_dir, PACKAGE_DATA_DIR "/images/tetravex/640x480");
					strcpy(op_env.objects_file, PACKAGE_DATA_DIR "/objects/tetravex-4x4-640x480");
				}
				if (ENG_SetOpEnvImageDir(op_env.image_dir)) printf("%s: %s\n", __func__, ENG_GetError());
				if (ENG_SetOpEnvObjectsFile(op_env.objects_file)) printf("%s: %s\n", __func__, ENG_GetError());
				if (ENG_LoadObjects(TRUE)) printf("%s: %s\n", __func__, ENG_GetError());
			} else if (event->object->id == REC_TETRAVEX_HISCORES) {
			} else if (event->object->id == REC_TETRAVEX_SOLVE) {
				/* Solve the game if not already solved */
				if (!solved) {
					for (count = 0; count < 9; count++) {
						/* Move the tiles into the away slots (no need to record the moves) */
						object = ENG_GetObject(LYR_TETRAVEX_TILES_00N + count * 4);
						target = ENG_GetObject(REC_TETRAVEX_SLOTS_00 + 9 + count);
						object->enabled = FALSE;	/* Disable the north tiles */
						/* Apply a move effect */
						effect.effect_id = LYR_TETRAVEX;
						effect.klasse = ENG_EFFECT_LINEAR_MOVE;
						effect.object = object;
						effect.frames = op_env.cycle_limit * TETRAVEX_TILE_MOVE_PERIOD + (rand() % 9);
						effect.paused = FALSE;
						effect.src_x = object->x; effect.src_y = object->y;
						effect.dst_x = target->x; effect.dst_y = target->y;
						if (ENG_CreateEffect(&effect)) printf("%s: %s\n", __func__, ENG_GetError());
					}
					/* Maximise the time which will stop the timer */
					timer_lastticks = 0;
					timer_time = TIMER_MAX;
					solved = TRUE;
				}
			} else if (event->object->id >= LYR_TETRAVEX_TILES_00N && event->object->id <= LYR_TETRAVEX_TILES_00N + 8 * 4) {
				/* Manage the dropping of the tiles */
				target = ENG_GetObjectUnderObject(ENG_GetObject(event->object->id + 1), ENG_HIT_POINT_CENTRE);	/* Underneath the middle of the east object */
				if (target->id >= REC_TETRAVEX_SLOTS_00 && target->id <= REC_TETRAVEX_SLOTS_00 + 8) {
					/* Drop the tile into the target home slot */
					event->object->x = target->x; event->object->y = target->y;
					/* printf("dropping %i into %i\n", event->object->id, target->id - REC_TETRAVEX_SLOTS_00); temp temp */
					for (count = 0; count < 18; count++) {
						if (tilelocations[count] == event->object->id) {
							tilelocations[count] = ENG_UNDEFINED;
							break;
						}
					}
					tilelocations[target->id - REC_TETRAVEX_SLOTS_00] = event->object->id;
				} else if (target->id >= REC_TETRAVEX_SLOTS_00 + 9 && target->id <= REC_TETRAVEX_SLOTS_00 + 17) {
					match = TRUE;
					/* Check northern edge */
					if ((target->id >= REC_TETRAVEX_SLOTS_00 + 12) &&
						(tilelocations[target->id - REC_TETRAVEX_SLOTS_00 - 3] != ENG_UNDEFINED && tilelocations[target->id - REC_TETRAVEX_SLOTS_00 - 3] != event->object->id) &&
						(ENG_GetObject(event->object->id + 0)->img_surface != ENG_GetObject(tilelocations[target->id - REC_TETRAVEX_SLOTS_00 - 3] + 2)->img_surface)) match = FALSE;
					/* Check eastern edge */
					if ((target->id != REC_TETRAVEX_SLOTS_00 + 11 && target->id != REC_TETRAVEX_SLOTS_00 + 14 && target->id != REC_TETRAVEX_SLOTS_00 + 17) &&
						(tilelocations[target->id - REC_TETRAVEX_SLOTS_00 + 1] != ENG_UNDEFINED && tilelocations[target->id - REC_TETRAVEX_SLOTS_00 + 1] != event->object->id) &&
						(ENG_GetObject(event->object->id + 1)->img_surface != ENG_GetObject(tilelocations[target->id - REC_TETRAVEX_SLOTS_00 + 1] + 3)->img_surface)) match = FALSE;
					/* Check southern edge */
					if ((target->id <= REC_TETRAVEX_SLOTS_00 + 14) &&
						(tilelocations[target->id - REC_TETRAVEX_SLOTS_00 + 3] != ENG_UNDEFINED && tilelocations[target->id - REC_TETRAVEX_SLOTS_00 + 3] != event->object->id) &&
						(ENG_GetObject(event->object->id + 2)->img_surface != ENG_GetObject(tilelocations[target->id - REC_TETRAVEX_SLOTS_00 + 3] + 0)->img_surface)) match = FALSE;
					/* Check western edge */
					if ((target->id != REC_TETRAVEX_SLOTS_00 + 9 && target->id != REC_TETRAVEX_SLOTS_00 + 12 && target->id != REC_TETRAVEX_SLOTS_00 + 15) &&
						(tilelocations[target->id - REC_TETRAVEX_SLOTS_00 - 1] != ENG_UNDEFINED && tilelocations[target->id - REC_TETRAVEX_SLOTS_00 - 1] != event->object->id) &&
						(ENG_GetObject(event->object->id + 3)->img_surface != ENG_GetObject(tilelocations[target->id - REC_TETRAVEX_SLOTS_00 - 1] + 1)->img_surface)) match = FALSE;
					if (match) {
						/* The target is valid so drop the tile into the away slot */
						event->object->x = target->x; event->object->y = target->y;
						/* printf("dropping %i into %i\n", event->object->id, target->id - REC_TETRAVEX_SLOTS_00); temp temp */
						for (count = 0; count < 18; count++) {
							if (tilelocations[count] == event->object->id) {
								tilelocations[count] = ENG_UNDEFINED;
								break;
							}
						}
						tilelocations[target->id - REC_TETRAVEX_SLOTS_00] = event->object->id;
						/* See if all the away slots are filled */
						solved = TRUE;
						for (count = 9; count < 18; count++) {
							if (tilelocations[count] == ENG_UNDEFINED) {
								solved = FALSE;
								break;
							}
						}					
						if (solved) {
							/* Disable the north tiles */
							for (count = 0; count < 9; count++) ENG_GetObject(LYR_TETRAVEX_TILES_00N + count * 4)->enabled = FALSE;
							/* Stop the timer */
							timer_state = TIMER_OFF;
							/* Check timer_time against hiscores */
							for (count = 0; count < HISCORES_MAX; count++) {
								if (timer_time < hiscore_time[count]) {
									/* Insert new hiscore */
									for (count2 = HISCORES_MAX - 1; count2 > count; count2--) {
										hiscore_time[count2] = hiscore_time[count2 - 1];
									}
									hiscore_time[count] = timer_time;
									/* Save tetravex_3x3.hiscore_time[] */
									for (count2 = 0; count2 < HISCORES_MAX; count2++) {
										sprintf(name, "tetravex_3x3.hiscore_time[%i]", count2 + 1);
										sprintf(value, "%i", hiscore_time[count2]);
										if (ENG_SetConfigOption(name, value)) printf("%s: %s\n", __func__, ENG_GetError());
									}
									/* Save tetravex_3x3.hiscore_last_position */
									sprintf(value, "%i", count + 1);
									if (ENG_SetConfigOption("tetravex_3x3.hiscore_last_position", value)) printf("%s: %s\n", __func__, ENG_GetError());
									printf("New hi-score :) position=%i\n", count + 1); //temp temp
									/* Change the event processor to tetravex_hiscores_new */
									op_env.event_processor = &tetravex_hiscores_new_event_processor;
									if (ENG_SetOpEnvEventProcessor(op_env.event_processor)) printf("%s: %s\n", __func__, ENG_GetError());
									/* Set the tetravex_hiscores_new image directory */
									/* Load the tetravex_hiscores_new objects file */
									if (op_env.screen_w < 640) {
										strcpy(op_env.image_dir, PACKAGE_DATA_DIR "/images/tetravex/320x240");
										strcpy(op_env.objects_file, PACKAGE_DATA_DIR "/objects/tetravex-hiscores-new-320x240");
									} else {
										strcpy(op_env.image_dir, PACKAGE_DATA_DIR "/images/tetravex/640x480");
										strcpy(op_env.objects_file, PACKAGE_DATA_DIR "/objects/tetravex-hiscores-new-640x480");
									}
									if (ENG_SetOpEnvImageDir(op_env.image_dir)) printf("%s: %s\n", __func__, ENG_GetError());
									if (ENG_SetOpEnvObjectsFile(op_env.objects_file)) printf("%s: %s\n", __func__, ENG_GetError());
									if (ENG_LoadObjects(TRUE)) printf("%s: %s\n", __func__, ENG_GetError());
									/* Hide the current layer */
									ENG_GetObject(LYR_TETRAVEX)->visible = FALSE;
									init_on_return = TRUE;
									break;
								}
							}
						}
					} else {
						/* The target is invalid so return the tile to its originating slot */
						for (count = 0; count < 18; count++) if (tilelocations[count] == event->object->id) break;
						target = ENG_GetObject(count + REC_TETRAVEX_SLOTS_00);
						/* Move the tile to its originating slot */
						effect.effect_id = LYR_TETRAVEX;
						effect.klasse = ENG_EFFECT_LINEAR_MOVE;
						effect.object = event->object;
						effect.frames = op_env.cycle_limit * TETRAVEX_TILE_MOVE_PERIOD;
						effect.paused = FALSE;
						effect.src_x = event->object->x; effect.src_y = event->object->y;
						effect.dst_x = target->x; effect.dst_y = target->y;
						if (ENG_CreateEffect(&effect)) printf("%s: %s\n", __func__, ENG_GetError());
					}				
				} else {
					/* The target is invalid so return the tile to its originating slot */
					for (count = 0; count < 18; count++) if (tilelocations[count] == event->object->id) break;
					target = ENG_GetObject(count + REC_TETRAVEX_SLOTS_00);
					/* Move the tile to its originating slot */
					effect.effect_id = LYR_TETRAVEX;
					effect.klasse = ENG_EFFECT_LINEAR_MOVE;
					effect.object = event->object;
					effect.frames = op_env.cycle_limit * TETRAVEX_TILE_MOVE_PERIOD;
					effect.paused = FALSE;
					effect.src_x = event->object->x; effect.src_y = event->object->y;
					effect.dst_x = target->x; effect.dst_y = target->y;
					if (ENG_CreateEffect(&effect)) printf("%s: %s\n", __func__, ENG_GetError());
				}
				/* Hide the tile shadows */
				shadoweast->visible = FALSE; shadowsouth->visible = FALSE;
			}
			break;
		case ENG_EVENT_QUITTING:
			break;
		case ENG_EVENT_POST_LOAD_OBJECTS:
			/* Store the objects_file so we know what to unload */
			strcpy(objects_file, ENG_GetObject(LYR_TETRAVEX)->objects_file);
			/* Store pointers to the tile image surfaces */
			for (count = 0; count < 9; count++) {
				tetravex_tile_images[count] = ENG_GetObject(LYR_TETRAVEX_TILES_00N + count * 4)->img_surface;
				if (count == 8) tetravex_tile_images[9] = ENG_GetObject(LYR_TETRAVEX_TILES_00N + count * 4)->img_over_surface;
			}
			/* Store pointers to the tile shadow objects */
			shadoweast = ENG_GetObject(REC_TETRAVEX_TILE_SHADOW_E);			/* East shadow */
			shadowsouth = ENG_GetObject(REC_TETRAVEX_TILE_SHADOW_E + 1);	/* South shadow */
			/* Store pointers to the timer objects and initialise the timer */
			timer1s = ENG_GetObject(REC_TETRAVEX_TIMER_1S);
			timer10s = ENG_GetObject(REC_TETRAVEX_TIMER_10S);
			timer1m = ENG_GetObject(REC_TETRAVEX_TIMER_1M);
			timer10m = ENG_GetObject(REC_TETRAVEX_TIMER_10M);
			timer_state = TIMER_OFF;
			/* Save tetravex.last_game_played */
			if (ENG_SetConfigOption("tetravex.last_game_played", "3x3")) printf("%s: %s\n", __func__, ENG_GetError());
			/* Retrieve tetravex_3x3.hiscore_time[] if possible */
			for (count = 0; count < HISCORES_MAX; count++) {
				sprintf(name, "tetravex_3x3.hiscore_time[%i]", count + 1);
				if (ENG_GetConfigOption(name, value)) {
					printf("%s: %s\n", __func__, ENG_GetError());
					hiscore_time[count] = TIMER_MAX;
				} else {
					if ((sscanf(value, "%i", &hiscore_time[count])) != 1) hiscore_time[count] = TIMER_MAX;
				}
				printf("%s: tetravex_3x3.hiscore_time[%i]=%i\n", __func__, count + 1, hiscore_time[count]); //temp temp
			}
			/* Start the 3x3 game by simulating clicking "3x3" */
			myevent.object = ENG_GetObject(REC_TETRAVEX_3X3);
			myevent.klasse = ENG_EVENT_CLICK;
			ENG_SendEvent(&myevent);
			break;
		case ENG_EVENT_UP:
			/* temp temp */
			op_env.viewport_y -= 4;
			//op_env.viewport_h -= 4;
			//op_env.viewport_screen_y -= 4;
			ENG_SetOpEnvViewport(op_env.viewport_x, op_env.viewport_y, op_env.viewport_w, op_env.viewport_h, op_env.viewport_screen_x, op_env.viewport_screen_y);
			break;
		case ENG_EVENT_UP_RIGHT:
			break;
		case ENG_EVENT_RIGHT:
			/* temp temp */
			op_env.viewport_x += 4;
			//op_env.viewport_w += 4;
			//op_env.viewport_screen_x += 4;
			ENG_SetOpEnvViewport(op_env.viewport_x, op_env.viewport_y, op_env.viewport_w, op_env.viewport_h, op_env.viewport_screen_x, op_env.viewport_screen_y);
			break;
		case ENG_EVENT_DOWN_RIGHT:
			break;
		case ENG_EVENT_DOWN:
			/* temp temp */
			op_env.viewport_y += 4;
			//op_env.viewport_h += 4;
			//op_env.viewport_screen_y += 4;
			ENG_SetOpEnvViewport(op_env.viewport_x, op_env.viewport_y, op_env.viewport_w, op_env.viewport_h, op_env.viewport_screen_x, op_env.viewport_screen_y);
			break;
		case ENG_EVENT_DOWN_LEFT:
			break;
		case ENG_EVENT_LEFT:
			/* temp temp */
			op_env.viewport_x -= 4;
			//op_env.viewport_w -= 4;
			//op_env.viewport_screen_x -= 4;
			ENG_SetOpEnvViewport(op_env.viewport_x, op_env.viewport_y, op_env.viewport_w, op_env.viewport_h, op_env.viewport_screen_x, op_env.viewport_screen_y);
			break;
		case ENG_EVENT_UP_LEFT:
			break;
		case ENG_EVENT_SELECT:
			break;
		case ENG_EVENT_SHUTDOWN:
			ENG_Quit();
			break;
		case ENG_EVENT_TOGGLE_FULLSCREEN:
			op_env.fullscreen = 1 - op_env.fullscreen;
			if (ENG_SetOpEnvFullScreen(op_env.fullscreen)) printf("%s: %s\n", __func__, ENG_GetError());
			break;
		case ENG_EVENT_MODIFIER1:
			/* temp temp */
			op_env.delay = 0;
			op_env.cycle_limit = 0;
			if (ENG_SetOpEnvDelayAndCycleLimit(op_env.delay, op_env.cycle_limit)) printf("%s: %s\n", __func__, ENG_GetError());
			break;
		case ENG_EVENT_MODIFIER2:
			/* temp temp */
			op_env.delay = 0;
			op_env.cycle_limit = 50;
			if (ENG_SetOpEnvDelayAndCycleLimit(op_env.delay, op_env.cycle_limit)) printf("%s: %s\n", __func__, ENG_GetError());
			break;
	}

}







